/*
 * drivers/serial/ns9xxx_serial.c
 *
 * Copyright (C) 2006 by Digi International Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */
/***********************************************************************
 *
 * !Revision: $Revision: 1.11 $
 * !Descr:  Serial Driver for NS9xxx
 * !Author: Markus Pietrek 
 *
 ***********************************************************************/
/*
 * !History:  07/011/19 : reserving physical address
 *                        reserving only GPIOs when needed
 */

#include <linux/config.h>
#include <linux/init.h>		/* __init */

#include <linux/module.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/console.h>	/* struct console */
#include <linux/ioport.h>	/* request_mem_region */

#include <linux/serial_core.h>	/* uart_console_device */

#include <asm/io.h>		/* writel */
#include <asm/delay.h>		/* udelay */

#include <asm/arch/regs-ser.h>
#include <asm/arch/regs-bbus.h>

#include <asm/arch/hardware.h>	/* ns9xxx_bbusclock() */
#include <asm/arch/irqs.h>
#include <asm/arch/system.h>



#define NR_PORTS	4
#define DRIVER_NAME	"ns9xxx_serial"

#define GPIO_TX		0
#define GPIO_RX		1
#define GPIO_RTS	2
#define GPIO_CTS	3
#define GPIO_DTR	4
#define GPIO_DSR	5
#define GPIO_RI		6
#define GPIO_DCD	7
#define GPIO_LAST	7

#ifdef CONFIG_SERIAL_NS9XXX_DEBUG
# define DEBUG_INIT		0x0001
# define DEBUG_MINOR		0x0002
# define DEBUG_ERROR_CRIT	0x8000

static int debug_level = DEBUG_ERROR_CRIT|DEBUG_INIT;

# define DEBUG_FN( FLG ) if( (debug_level & (FLG)) == (FLG) ) \
		printk("\r%s:line %d\n", __FUNCTION__, __LINE__);
#else /* CONFIG_SERIAL_NS9XXX_DEBUG */
# define DEBUG_FN( n )
#endif /* CONFIG_SERIAL_NS9XXX_DEBUG */

#ifdef CONFIG_SERIAL_NS9750_HANDSHAKE
# define HANDSHAKE 1
#else
# define HANDSHAKE 0
#endif  /* CONFIG_SERIAL_NS9750_HANDSHAKE */

/* Individual ports can be disabled by kernel configuration. Create the makros
 * to enable/disable */
#define CFG_PORT_STATUS(x) disabled : x

#ifdef CONFIG_SERIAL_NS9XXX_ENABLE_PORT_0
# define CFG_PORT_STATUS_0 CFG_PORT_STATUS(0)
#else
# define CFG_PORT_STATUS_0 CFG_PORT_STATUS(1)
#endif
#ifdef CONFIG_SERIAL_NS9XXX_ENABLE_PORT_1
# define CFG_PORT_STATUS_1 CFG_PORT_STATUS(0)
#else
# define CFG_PORT_STATUS_1 CFG_PORT_STATUS(1)
#endif
#ifdef CONFIG_SERIAL_NS9XXX_ENABLE_PORT_2
# define CFG_PORT_STATUS_2 CFG_PORT_STATUS(0)
#else
# define CFG_PORT_STATUS_2 CFG_PORT_STATUS(1)
#endif
#ifdef CONFIG_SERIAL_NS9XXX_ENABLE_PORT_3
# define CFG_PORT_STATUS_3 CFG_PORT_STATUS(0)
#else
# define CFG_PORT_STATUS_3 CFG_PORT_STATUS(1)
#endif

typedef struct
{
	struct uart_port       	port;
	int 		        hw_port;/* serial port to use. -1 for disabled */
	char 		        hw_handshake;/* 1 for enabled, 0 for disabled */
	spinlock_t		lock; /* for ser_tx_chars */
	const char*	       	gpios;
        char                    disabled;
} ns_ser_port;

ns_ser_port ns_ser_ports[ NR_PORTS ] = {
	{ hw_port: -1, CFG_PORT_STATUS_0 },
	{ hw_port: -1, CFG_PORT_STATUS_1 },
	{ hw_port: -1, CFG_PORT_STATUS_2 },
	{ hw_port: -1, CFG_PORT_STATUS_3 }
};

static int __init ns_ser_init( void );
static void __init ns_ser_init_ports( int hw_port );
static void __exit ns_ser_exit( void );

static unsigned int ns_ser_tx_empty( struct uart_port* port );
static void ns_ser_set_mctrl( struct uart_port* port, unsigned int mctrl );
static unsigned int ns_ser_get_mctrl( struct uart_port* port );
static void ns_ser_stop_tx( struct uart_port* port );
static void ns_ser_start_tx( struct uart_port* port );
static void ns_ser_stop_rx( struct uart_port* port );
static void ns_ser_enable_ms( struct uart_port* port );
static void ns_ser_break_ctl( struct uart_port* port, int break_state );
static int ns_ser_startup( struct uart_port* port );
static inline void ns_ser_wait_for_fifo_empty( ns_ser_port* sport );
static inline void ns_ser_wait_for_xmitr( ns_ser_port* sport );
static void ns_ser_shutdown( struct uart_port* port );
static void ns_ser_set_termios( struct uart_port* port,
				struct termios* termios,
				struct termios* old );
static void ns_ser_pm( struct uart_port* port, unsigned int state,
			   unsigned int oldstate );
static const char* ns_ser_type( struct uart_port* port );
static void ns_ser_release_port( struct uart_port* port );
static int ns_ser_request_port( struct uart_port* port );
static void ns_ser_config_port( struct uart_port* port, int flags );
static int ns_ser_verify_port( struct uart_port* port,
			       struct serial_struct* ser );
static irqreturn_t ns_ser_tx_int( int irq, void* dev_id,
				  struct pt_regs *regs );
static irqreturn_t ns_ser_rx_int( int irq, void* dev_id,
				  struct pt_regs *regs );
static void ns_ser_rx_chars( ns_ser_port* sport, struct pt_regs* regs );
static void ns_ser_tx_chars( ns_ser_port* sport );
static inline void ns_ser_tx_char( ns_ser_port* sport, char c );
static void ns_ser_set_bitrate( int cflag, int quot, int hw_port );

static int ns_console_cbaud = B9600;

#ifdef CONFIG_SERIAL_NS9750_CONSOLE

static int __init ns_console_init( void );
static void __init ns_console_get_options( int* baud, int* parity, int* bits,
					   int hw_port );
static int __init ns_console_setup( struct console* co, char* options );
static void ns_console_write( struct console* co, const char* s,
			      unsigned int count );

static struct uart_driver ns_ser_reg;
static struct console ns_console = {
	.name		= "ttyS",
	.write		= ns_console_write,
	.device		= uart_console_device,
	.setup		= ns_console_setup,
	.flags		= CON_PRINTBUFFER,
	.index		= -1,
	.data		= &ns_ser_reg
};

# define SERIAL_NS_CONSOLE	&ns_console
#else
# define SERIAL_NS_CONSOLE	NULL

#endif /* CONFIG_SERIAL_NS9750_CONSOLE */

static struct uart_driver ns_ser_reg = {
	.owner			= THIS_MODULE,
	.driver_name		= DRIVER_NAME,
	.devfs_name		= "ttyS/",
	.dev_name		= "ttyS",
	.major			= TTY_MAJOR,
	.minor			= 64,
	.nr			= NR_PORTS,
	.cons			= SERIAL_NS_CONSOLE
};

static struct uart_ops ns_ser_pops = {
	.tx_empty	= ns_ser_tx_empty,
	.set_mctrl	= ns_ser_set_mctrl,
	.get_mctrl	= ns_ser_get_mctrl,
	.stop_tx	= ns_ser_stop_tx,
	.start_tx	= ns_ser_start_tx,
	.stop_rx	= ns_ser_stop_rx,
	.enable_ms	= ns_ser_enable_ms,
	.break_ctl	= ns_ser_break_ctl,
	.startup	= ns_ser_startup,
	.shutdown	= ns_ser_shutdown,
	.set_termios	= ns_ser_set_termios,
	.pm		= ns_ser_pm,
	.type		= ns_ser_type,
	.release_port	= ns_ser_release_port,
	.request_port	= ns_ser_request_port,
	.config_port	= ns_ser_config_port,
	.verify_port	= ns_ser_verify_port,
};
/* order of GPIOs is GPIO_TX to GPIO_LAST, used for ns_ser_port.gpios  */
static const char ns_ser_gpios[][ GPIO_LAST + 1 ] = {
	{  0,  1,  2,  3,  4,  5,  6,  7 }, /* Port B */
	{  8,  9, 10, 11, 12, 13, 14, 15 }, /* Port A */
	{ 40, 41, 42, 43, 20, 21, 22, 23 }, /* Port C */
	{ 44, 45, 46, 47, 24, 25, 26, 27 }  /* Port D */
};

static inline u32 serial_in(ns_ser_port *sport, u32 reg)
{
	return readl(sport->port.membase + reg);
}

static inline void serial_out(ns_ser_port *sport, u32 reg, u32 val)
{
	writel(val, sport->port.membase + reg);
}

static int __init ns_ser_init( void )
{
	int ret, i;
	ns_ser_port* sport;

	printk( KERN_INFO "Serial: NS9xx0 driver $Revision: 1.11 $\n" );

	/* @TODO: read ports masked from command line */
	for( i = 0; i < NR_PORTS; i++ )
		ns_ser_init_ports( i );

	ret = uart_register_driver( &ns_ser_reg );
	if( ret < 0 )
		return ret;

	ns_ser_reg.tty_driver->init_termios.c_cflag =
		(ns_ser_reg.tty_driver->init_termios.c_cflag & ~CBAUD) |
		ns_console_cbaud;

	for( i = 0; i < NR_PORTS; i++ ) {
		if( ns_ser_ports[ i ].hw_port != -1 ) {
			sport = &ns_ser_ports[i];
                        if( !sport->disabled ) {
                                uart_add_one_port( &ns_ser_reg, &sport->port );
                                if (!(serial_in(sport, SER_CTRLA)  & NS_SER_CTRL_A_CE)) {
                                        serial_out(sport, SER_CTRLA, NS_SER_CTRL_A_CE);
                                }
                        } else {
                                printk( KERN_INFO DRIVER_NAME ": Port %i disabled in kernel config\n", i );
                        }
		}
	}
	return ret;
}

static void __init ns_ser_init_ports( int hw_port )
{
	ns_ser_port* sport = &ns_ser_ports[ hw_port ];

	if( sport->hw_port != -1 )
		/* already initialized */
		return;

	sport->hw_port        = hw_port;
	sport->port.uartclk   = ns9xxx_bbusclock();
	sport->port.fifosize  = 32;
	sport->port.line      = hw_port;
	sport->port.iotype    = UPIO_MEM;
	sport->port.ops       = &ns_ser_pops;
	sport->port.membase   = (void __iomem *)(SER_MODULE_BASE_VA+SER_CH_OFF(hw_port));
	sport->port.mapbase   = SER_MODULE_BASE_PA+SER_CH_OFF(hw_port);
	sport->port.flags     = UPF_BOOT_AUTOCONF;
	
	/* note only rhw_port irq. tx is +1 */
	sport->port.irq       = IRQ_BBUS_SERBRX + (2*hw_port);
	sport->gpios          = ns_ser_gpios[ hw_port ];
	


    /* IK-FNA: Make Handshake only work on port 1*/
	 if (hw_port == 1)
   {
     printk( KERN_INFO DRIVER_NAME ": Port hw_port %i | Handshake = %i \n", hw_port, HANDSHAKE );
     sport->hw_handshake = HANDSHAKE;
   }
   else
   {
     printk( KERN_INFO DRIVER_NAME ": Port hw_port %i | Handshake = 0 \n", hw_port );
     sport->hw_handshake = 0;
   }

	
	
	spin_lock_init( &sport->port.lock);
	spin_lock_init( &sport->lock );
}


static void __exit ns_ser_exit( void )
{
	int i;
	ns_ser_port* sport = NULL;
        
	for( i = 0; i < NR_PORTS; i++ ) {
                sport = &ns_ser_ports[ i ];
                
		if( sport->hw_port != -1 ) {
                        if( !sport->disabled )
                                uart_remove_one_port( &ns_ser_reg,
                                                      &sport->port );
                }
        }
        
	uart_unregister_driver( &ns_ser_reg );
}

static unsigned int ns_ser_tx_empty( struct uart_port* port )
{
	ns_ser_port* sport = (ns_ser_port*) port;

	DEBUG_FN( DEBUG_MINOR );

	return( (serial_in(sport, SER_STATUSA) & NS_SER_STAT_A_TRDY) ? TIOCSER_TEMT : 0 );
}

static void ns_ser_set_mctrl( struct uart_port* port, unsigned int mctrl )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	unsigned int mcr;

	DEBUG_FN(DEBUG_INIT);

	mcr = serial_in(sport, SER_CTRLA);

	if (mctrl & TIOCM_RTS)
		mcr &= ~NS_SER_CTRL_A_RTS;
	else
		mcr |= NS_SER_CTRL_A_RTS;

	if (mctrl & TIOCM_DTR)
		mcr &= ~NS_SER_CTRL_A_DTR;
	else
		mcr |= NS_SER_CTRL_A_DTR;

	serial_out(sport, SER_CTRLA, mcr);
}

static unsigned int ns_ser_get_mctrl( struct uart_port* port )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	unsigned int result = 0;
	unsigned int status;

	DEBUG_FN( DEBUG_INIT );

	status = serial_in(sport, SER_STATUSA);

	if (!(status & NS_SER_STAT_A_DCD))
		result |= TIOCM_CD;
	if (!(status & NS_SER_STAT_A_CTS))
		result |= TIOCM_CTS;
	if (!(status & NS_SER_STAT_A_DSR))
		result |= TIOCM_DSR;
	if (!(status & NS_SER_STAT_A_RI))
		result |= TIOCM_RI;

	return result;
}

static void ns_ser_stop_tx( struct uart_port* port )
{
	u32 ctrl_a;
	ns_ser_port* sport = (ns_ser_port*) port;

	ctrl_a = serial_in(sport, SER_CTRLA ) & ~NS_SER_CTRL_A_TIC_HALF;
	serial_out(sport, SER_CTRLA, ctrl_a);
}

/***********************************************************************
 * @Function: ns_ser_start_tx
 * @Return: n/a
 * @Descr: enables transmit interrupt on FIFO empty
 ***********************************************************************/
static void ns_ser_start_tx( struct uart_port* port )
{
	u32 ctrl_a;
	ns_ser_port* sport = (ns_ser_port*) port;

	ctrl_a = serial_in(sport, SER_CTRLA ) | NS_SER_CTRL_A_TIC_HALF;
	serial_out(sport, SER_CTRLA, ctrl_a);
}

static void ns_ser_stop_rx( struct uart_port* port )
{
	u32 ctrl_a;
	ns_ser_port* sport = (ns_ser_port*) port;

	DEBUG_FN( DEBUG_INIT );

	ctrl_a = serial_in(sport, SER_CTRLA ) & ~NS_SER_CTRL_A_RIE_RDY & ~NS_SER_CTRL_A_RIE_CLOSED;
	serial_out(sport, SER_CTRLA, ctrl_a);
}

static void ns_ser_enable_ms( struct uart_port* port )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	unsigned int status;

	DEBUG_FN( DEBUG_INIT );

	status = serial_in(sport, SER_STATUSA);

	/* XXX */
	/* status |= NS_SER_CTRL_A_RIC_MA; */

	serial_out(sport, SER_STATUSA, status);
}

static void ns_ser_break_ctl( struct uart_port* port, int break_state )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	u32 brk;

	DEBUG_FN(DEBUG_INIT);

	brk = serial_in(sport, SER_CTRLA);

	if (break_state == -1)
		brk |= NS_SER_CTRL_A_BRK;  /* Assert break */
	else
		brk &= ~NS_SER_CTRL_A_BRK;  /* Deassert break */

	serial_out(sport, SER_CTRLA, brk);
}

/***********************************************************************
 * @Function: ns_ser_startup
 * @Return: 0 if ok otherwise != 0 if interrupts could not be registered
 * @Descr:
 ***********************************************************************/

static int ns_ser_startup( struct uart_port* port )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	u32 ctrl_a;
	int ret = -EBUSY;
        int i = 0;

	DEBUG_FN( DEBUG_INIT );

	/* request area in ser_module */
	if( request_mem_region( sport->port.mapbase, SER_CH_OFFSET1,
				DRIVER_NAME ) == NULL) {
		printk( KERN_ERR DRIVER_NAME ":request_mem_region failed\n" );

		goto error_region;
	}

	/* request gpios */
        
	for( i = 0; i < GPIO_LAST; i++ )
		if( i <= GPIO_RX || sport->hw_handshake )
			if(ns9xxx_request_gpio(sport->gpios[i], DRIVER_NAME)!=0) {
				printk(	KERN_ERR DRIVER_NAME
					":GPIO %i reserved\n", sport->gpios[i] );
				goto error_gpio;
			}

	/* configure GPIOS. we need at least tx and rx */

        ns9xxx_gpio_cfgpin( sport->gpios[ GPIO_TX ], GPIO_CFG_FUNC_0 );
        ns9xxx_gpio_cfgpin( sport->gpios[ GPIO_RX ], GPIO_CFG_FUNC_0 );
                
	if( sport->hw_handshake ) {
		ns9xxx_gpio_cfgpin( sport->gpios[ GPIO_RTS ], GPIO_CFG_FUNC_0 );
		ns9xxx_gpio_cfgpin( sport->gpios[ GPIO_CTS ], GPIO_CFG_FUNC_0 );
		ns9xxx_gpio_cfgpin( sport->gpios[ GPIO_DTR ], GPIO_CFG_FUNC_0 );
		ns9xxx_gpio_cfgpin( sport->gpios[ GPIO_DSR ], GPIO_CFG_FUNC_0 );
		ns9xxx_gpio_cfgpin( sport->gpios[ GPIO_RI  ], GPIO_CFG_FUNC_0 );
		ns9xxx_gpio_cfgpin( sport->gpios[ GPIO_DCD ], GPIO_CFG_FUNC_0 );
	}
        
	ret = request_irq( sport->port.irq, ns_ser_rx_int, 0,
			   DRIVER_NAME " rx", sport );
	if( ret )
                goto error_rx;

	ret = request_irq( sport->port.irq+1, ns_ser_tx_int, 0,
			   DRIVER_NAME " tx", sport );
	if( ret )
                goto error_tx;

	spin_lock( &sport->port.lock );

	if( serial_in(sport, SER_CTRLA) & NS_SER_CTRL_A_CE )
		/* this is a console if already enabled, maybe there is still
		 * something to be printed. wait until modified */
		ns_ser_wait_for_fifo_empty(sport);

	/* enable receive interrupts only */
	ctrl_a = serial_in(sport, SER_CTRLA ) | NS_SER_CTRL_A_CE | NS_SER_CTRL_A_RIE_RDY | NS_SER_CTRL_A_RIE_CLOSED;
	serial_out(sport, SER_CTRLA, ctrl_a);

	serial_out(sport, SER_CTRLB, NS_SER_CTRL_B_RCGT);
	spin_unlock( &sport->port.lock );

	/* Enable modem status interrupts */
	spin_lock_irq( &sport->port.lock );
	ns_ser_enable_ms( &sport->port );
	spin_unlock_irq( &sport->port.lock );

	/* termios settings will be configured later by set_termios */

        return 0;

error_tx:
	free_irq( sport->port.irq, sport );

error_rx:
error_gpio:
        /* we may have some reserved gpio's release them */
	/* i is +1 of the last reserved gpio */
	for( i--; i > 0; i-- )
		ns9xxx_release_gpio( sport->gpios[ i ] );

	release_mem_region( sport->port.mapbase, SER_CH_OFFSET1 );

error_region:
	return ret;
}

/***********************************************************************
 * @Function: ns_ser_wait_for_fifo_empty
 * @Return: nothing
 * @Descr: doesn't lock for port accesses, must been done by calling function
 ***********************************************************************/

static inline void ns_ser_wait_for_fifo_empty( ns_ser_port* sport )
{
	while( !(serial_in(sport, SER_STATUSA) & NS_SER_STAT_A_TEMPTY) )
		/* do nothing, wait for characters in FIFO being sent */
		udelay( 1 );

	/* @TODO: until we have buffer closed again, do just a delay, enough
         * for 38400 baud */
	udelay( 1000 );
}

static inline void ns_ser_wait_for_xmitr( ns_ser_port* sport )
{
	while( !(serial_in(sport, SER_STATUSA) & NS_SER_STAT_A_TRDY) )
		/* do nothing, wait for characters in FIFO being sent */
		udelay( 1 );
}

static void ns_ser_shutdown( struct uart_port* port )
{
	ns_ser_port* sport = (ns_ser_port*) port;
        int i;
	u32 ctrl_a;
	unsigned long flags;

	DEBUG_FN( DEBUG_INIT );

	/* disable serial port */

	ns_ser_wait_for_fifo_empty(sport);
	/* @TODO: printk could write in this small time to FIFO */

	spin_lock_irqsave( &sport->port.lock, flags );

	ctrl_a = serial_in(sport, SER_CTRLA ) & ~NS_SER_CTRL_A_RIE_RDY;
	serial_out(sport, SER_CTRLA, ctrl_a);

	spin_unlock_irqrestore( &sport->port.lock, flags );

	free_irq( sport->port.irq, sport );
	free_irq( sport->port.irq + 1, sport );	/* transmit */

	/* release all gpios  */

	for( i = 0; i < GPIO_LAST; i++ )
		if( i <= GPIO_RX || sport->hw_handshake )
			ns9xxx_release_gpio( sport->gpios[ i ] );

	/* release area in ser_module */
	release_mem_region( sport->port.mapbase, SER_CH_OFFSET1 ); 
}

static void ns_ser_set_termios( struct uart_port* port,
				struct termios* termios,
				struct termios* old )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	unsigned long flags;
	unsigned int baud, quot;

	DEBUG_FN( DEBUG_INIT );

	/* Ask the core to calculate the divisor for us. */
	baud = uart_get_baud_rate( port, termios, old, 0, port->uartclk/16 );
	quot = uart_get_divisor( port, baud );

	spin_lock_irqsave( &sport->port.lock, flags );
	uart_update_timeout( port, termios->c_cflag, baud );

	ns_ser_wait_for_fifo_empty(sport);
	ns_ser_set_bitrate( termios->c_cflag, quot, sport->hw_port );

	if( UART_ENABLE_MS( &sport->port, termios->c_cflag ) )
		ns_ser_enable_ms( &sport->port );

	spin_unlock_irqrestore( &sport->port.lock, flags );
}

static void ns_ser_pm( struct uart_port* port, unsigned int state,
		       unsigned int oldstate )
{
	/* power management not supported yet */
}

static const char* ns_ser_type( struct uart_port* port )
{
	return "NS9xx0";
}

static void ns_ser_release_port( struct uart_port* port )
{
	DEBUG_FN( DEBUG_INIT );
}

static int ns_ser_request_port( struct uart_port* port )
{
	DEBUG_FN( DEBUG_INIT );

	return 0;
}

static void ns_ser_config_port( struct uart_port* port, int flags )
{
	ns_ser_port* sport = (ns_ser_port*) port;

	DEBUG_FN( DEBUG_INIT );

	if( flags & UART_CONFIG_TYPE && ns_ser_request_port(&sport->port)==0)
		sport->port.type = PORT_NS9750;
}

static int ns_ser_verify_port( struct uart_port* port,
				   struct serial_struct* ser )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	int ret = 0;

	DEBUG_FN( DEBUG_INIT );

	if( ser->type != PORT_UNKNOWN && ser->type != PORT_NS9750 )
		ret = -EINVAL;
	if( sport->port.irq != ser->irq )
		ret = -EINVAL;
	if( ser->io_type != UPIO_MEM )
		ret = -EINVAL;

	return ret;
}

static irqreturn_t ns_ser_tx_int( int irq, void* dev_id,
				  struct pt_regs *regs )
{
	ns_ser_port* sport = (ns_ser_port*) dev_id;

	spin_lock( &sport->lock );
	ns_ser_tx_chars( sport );
	spin_unlock( &sport->lock );

	return IRQ_HANDLED;
}

static irqreturn_t ns_ser_rx_int( int irq, void* dev_id,
				  struct pt_regs *regs )
{
	ns_ser_port* sport = (ns_ser_port*) dev_id;

	spin_lock( &sport->lock );
	ns_ser_rx_chars( sport, regs );
	spin_unlock( &sport->lock );

	return IRQ_HANDLED;
}

/***********************************************************************
 * @Function: ns_ser_rx_chars
 * @Return: n/a
 * @Descr: receives all characters from serial FIFO and stores them in tty. Is
 *         called by interrupt handler.
 ***********************************************************************/

static void ns_ser_rx_chars( ns_ser_port* sport, struct pt_regs* regs )
{
	unsigned int stat;
	unsigned int available;
	unsigned int char_cache;
	char c;
	unsigned int flag;
	struct tty_struct *tty = sport->port.info->tty;

	stat = serial_in(sport, SER_STATUSA);

        if( stat & NS_SER_STAT_A_RBC ) {
		serial_out( sport, SER_STATUSA, NS_SER_STAT_A_RBC );
                stat = serial_in(sport, SER_STATUSA);
        }

	while( 1 ) {
		/* read 0-4 characters */
		if ( ! (stat & NS_SER_STAT_A_RRDY ) )
			/* no more characters, return */
			break;

		available = (stat & NS_SER_STAT_A_RXFDB_MA) >> 20;
		if ( !available )
			available = 4;

		char_cache = serial_in( sport, SER_FIFO);
                
    //printk( KERN_INFO DRIVER_NAME ": stat: %08X \n", stat);

		flag = TTY_NORMAL;
		if ( stat & NS_SER_STAT_A_ROVER ) {
			sport->port.icount.overrun++;
			flag = TTY_OVERRUN;
		} else if( stat & NS_SER_STAT_A_RFE ) {
			sport->port.icount.frame++;
			flag = TTY_FRAME;
		} else if( stat & NS_SER_STAT_A_RPE ) {
			sport->port.icount.parity++;
			flag = TTY_PARITY;
		} else if( stat & NS_SER_STAT_A_RBRK ){
			sport->port.icount.brk++;
			flag = TTY_BREAK;
		}

		/* characters read from FIFO, give them kernel */

		while( available ) {

			available--;
			c = char_cache & 0xff;
			char_cache >>= 8;

			sport->port.icount.rx++;

			tty_insert_flip_char(tty, c, flag);
		}

		/* check for next characters */
		stat = serial_in(sport, SER_STATUSA);

		stat = serial_in(sport, SER_STATUSA);
	}

        if( !tty->low_latency )
                tty_flip_buffer_push( tty );

#ifdef SUPPORT_SYSRQ
	sport->port.sysrq = 0;
#endif
}

/***********************************************************************
 * @Function: ns_ser_tx_char
 * @Return: nothing
 * @Descr: sends one character (in interrupt handler
 ***********************************************************************/

inline static void ns_ser_tx_char( ns_ser_port* sport, char c )
{
	writeb(c, sport->port.membase + SER_FIFO);
}

/***********************************************************************
 * @Function: ns_tx_chars
 * @Return: n/a
 * @Descr: sends as many characters to FIFO as possible.
 ***********************************************************************/

static void ns_ser_tx_chars( ns_ser_port* sport )
{
	struct circ_buf* xmit = &sport->port.info->xmit;

	if( sport->port.x_char ) {
		ns_ser_tx_char( sport, sport->port.x_char );
		sport->port.icount.tx++;
		sport->port.x_char = 0;
        } else if( !uart_circ_empty( xmit ) && !uart_tx_stopped( &sport->port ) ) {
                /* transmit characters */
                while( serial_in(sport, SER_STATUSA) & NS_SER_STAT_A_TRDY ) {
                        ns_ser_tx_char( sport, xmit->buf[ xmit->tail ] );
                        xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
                        sport->port.icount.tx++;
                        if( uart_circ_empty( xmit ) )
                                break;
                }

                if( uart_circ_chars_pending( xmit ) < WAKEUP_CHARS )
                        uart_write_wakeup( &sport->port );
        }

        /* uart may be empty now */
        if( uart_circ_empty( xmit ) || uart_tx_stopped( &sport->port ) )
                /* wait until they are sent */
                ns_ser_stop_tx( &sport->port );
}

static void ns_ser_set_bitrate( int cflag, int quot, int hw_port )
{
	ns_ser_port* sport = &ns_ser_ports[hw_port];
	u32 ctrl_a, tmp, bitrate;

	ctrl_a = NS_SER_CTRL_A_CE;
	/* byte size and parity */
	switch( cflag & CSIZE )
	{
	    case CS5: ctrl_a |= NS_SER_CTRL_A_WLS_5; break;
	    case CS6: ctrl_a |= NS_SER_CTRL_A_WLS_6; break;
	    case CS7: ctrl_a |= NS_SER_CTRL_A_WLS_7; break;
	    case CS8: ctrl_a |= NS_SER_CTRL_A_WLS_8; break;
	}
	if( cflag & CSTOPB )
		ctrl_a |= NS_SER_CTRL_A_STOP;
	if( cflag & PARENB )
		ctrl_a |= NS_SER_CTRL_A_PE;
	if( !(cflag & PARODD) )
		ctrl_a |= NS_SER_CTRL_A_EPS;
	/* for PARODD, no value has to be set */

	bitrate = ( NS_SER_BITRATE_EBIT |
		    NS_SER_BITRATE_CLKMUX_BCLK |
		    NS_SER_BITRATE_TMODE |
		    NS_SER_BITRATE_TCDR_16 |
		    NS_SER_BITRATE_RCDR_16 |
		    ( ( quot - 1 ) &
		      NS_SER_BITRATE_N_MA ) );

	tmp = serial_in(sport, SER_CTRLA) & ~(NS_SER_CTRL_A_WLS_8 | NS_SER_CTRL_A_STOP | NS_SER_CTRL_A_PE | NS_SER_CTRL_A_EPS);
	tmp |= ctrl_a;
	serial_out(sport, SER_CTRLA, tmp | ctrl_a);

	serial_out(sport, SER_BITRATE, bitrate);

	/* @TODO: when chip supports it, do real calculation */
	serial_out(sport, SER_RXCHARTIMER, NS_SER_RX_CHAR_TIMER_TRUN);
}

#ifdef CONFIG_SERIAL_NS9750_CONSOLE

struct baud_rates {
	unsigned int rate;
	unsigned int cflag;
};

static struct baud_rates baud_rates[] = {
	{ 921600, B921600 },
	{ 460800, B460800 },
	{ 230400, B230400 },
	{ 115200, B115200 },
	{  57600, B57600  },
	{  38400, B38400  },
	{  19200, B19200  },
	{   9600, B9600   },
	{   4800, B4800   },
	{   2400, B2400   },
	{   1200, B1200   },
	{      0, B38400  }
};

static int __init ns_console_init( void )
{
	register_console( &ns_console );
	return 0;
}

/*
 * If the port was already initialised (eg, by a boot loader),
 * try to determine the current setup.
 */
static void __init ns_console_get_options( int* baud, int* parity, int* bits,
					   int hw_port)
{
	ns_ser_port* sport = &ns_ser_ports[hw_port];
	unsigned int ctrl_a;
	unsigned int bitrate;

	/* no need to worry about other processes accessing bitrate */
	ctrl_a = serial_in(sport, SER_CTRLA);
	bitrate = serial_in(sport, SER_BITRATE);

	if( !(ctrl_a & NS_SER_CTRL_A_CE ) )
		/* not enabled, don't modify defaults */
		return;

	switch( ctrl_a & NS_SER_CTRL_A_WLS_MA ) {
	    case NS_SER_CTRL_A_WLS_5 : *bits = 5; break;
	    case NS_SER_CTRL_A_WLS_6 : *bits = 6; break;
	    case NS_SER_CTRL_A_WLS_7 : *bits = 7; break;
	    case NS_SER_CTRL_A_WLS_8 :
	    default: *bits = 8; break;
	}

	switch( ctrl_a & NS_SER_CTRL_A_PE ) {
	    case NS_SER_CTRL_A_PE : *parity = 'e'; break;
	    case (NS_SER_CTRL_A_PE|NS_SER_CTRL_A_EPS) : *parity = 'o'; break;
	    default: *parity = 'n'; break;
	}

	/* expects a clk divide rate of 16, no check is done yet */
	*baud = ns9xxx_bbusclock()/(((bitrate & NS_SER_BITRATE_N_MA) + 1)*16);
}

static int __init ns_console_setup( struct console* co, char* options )
{
	ns_ser_port* sport;
	int baud = 9600;
	int bits = 8;
	int parity = 'n';
	int flow = 'n';
	int cflag;
	int ret;
	int i;

	/* only accept supported UARTs, otherwise take first */
	if( co->index == -1 || co->index >= NR_PORTS )
		co->index = 0;

	ns_ser_init_ports( co->index );

	sport = &ns_ser_ports[ co->index ];

	if( options )
		uart_parse_options( options, &baud, &parity, &bits, &flow );
	else
		ns_console_get_options( &baud, &parity, &bits, co->index );

	/* create cflag as co->cflag may not be set if no options */
	cflag = 0;
	switch( bits )
	{
	    case 5: cflag |= CS5; break;
	    case 6: cflag |= CS6; break;
	    case 7: cflag |= CS7; break;
	    case 8:
	    default: cflag |= CS8; break;
	}
	switch( parity ) {
	    case 'o':
	    case 'O':
		cflag |= PARODD;
		/*fall through*/
	    case 'e':
	    case 'E':
		cflag |= PARENB;
		break;
	}

	ns9xxx_gpio_cfgpin( ns_ser_gpios[ co->index ][ GPIO_TX ], GPIO_CFG_FUNC_0 );

	/* no need to worry about other processes accessing bitrate */
	ns_ser_set_bitrate( cflag, ns9xxx_bbusclock() / ( baud * 16 ), co->index );

	ret = uart_set_options( &sport->port, co, baud, parity, bits, flow );

	/* remember baudrate for termios. this will fail if multiple consoles
	 * are used for this driver with different baudrates, only the last one
	 * will be used for all */
	for( i = 0; baud_rates[i].rate; i++ )
		if( baud_rates[i].rate <= baud )
			break;

	ns_console_cbaud = baud_rates[i].cflag;

	/* for default consoles in kernel commandline, co is not initialized */
	return ret;
}

/*
 * Interrupts are disabled on entering
 */
static void ns_console_write( struct console* co, const char* s,
			      unsigned int count )
{
	ns_ser_port* sport = &ns_ser_ports[co->index];
	char c = *s;
	u32 ctrl_old, ctrl;

	/* save state and disable interrupts */
	/* we keep the baudrate settings of /dev/ttyS0 */
	ctrl_old = serial_in(sport, SER_CTRLA);
	/* may be disabled, so renable it  */
	ctrl = (ctrl_old | NS_SER_CTRL_A_CE) & ~NS_SER_CTRL_A_TIC_MA &
		~NS_SER_CTRL_A_RIC_MA & ~NS_SER_CTRL_A_RIE_MA;
	serial_out(sport, SER_CTRLA, ctrl);

	while( count ) {
		ns_ser_wait_for_xmitr(sport);

		if( c == '\n' ) {
                        writeb( '\r', sport->port.membase + SER_FIFO);
                        ns_ser_wait_for_xmitr(sport);
                }
                
                writeb( c, sport->port.membase + SER_FIFO);

                /* next character */
                count--;
                s++;
                c = *s;
	}

	if( !(ctrl_old & NS_SER_CTRL_A_CE ) )
		/* send whole message  */
		ns_ser_wait_for_fifo_empty(sport);

	/* enable interrupts */
	serial_out(sport, SER_CTRLA, ctrl_old);
}

console_initcall( ns_console_init );
#endif /* CONFIG_SERIAL_NS9750_CONSOLE */

module_init( ns_ser_init );
module_exit( ns_ser_exit );

MODULE_AUTHOR( "Digi International Inc" );
MODULE_DESCRIPTION( "NS9xxx serial driver $Revision: 1.11 $" );
MODULE_LICENSE( "GPL" );
