/*
 * linux/drivers/i2c/busses/i2c-ns9xxx.c
 *
 * Copyright (C) 2006 by Digi International Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <linux/i2c.h>

#include <asm/arch/irqs.h>
#include <asm/arch/system.h>
#include <asm/arch/hardware.h>
#include <asm/arch/processor.h>
#include <asm/arch/regs-i2c.h>
#include <asm/arch/regs-bbus.h>

#define DRIVER_NAME			"i2c-ns9xxx"
#define NS_BBUS_MASTER_RESET_I2C	(0x00000080)

#if defined(CONFIG_I2C_CLK_SPEED_NORMAL)
# define I2C_CLK_FREQ	(100000)
# define I2C_SCL_DELAY	(3)
#elif defined(CONFIG_I2C_CLK_SPEED_FAST)
# define I2C_CLK_FREQ	(400000)
# define I2C_SCL_DELAY	(24)
#else
# define I2C_CLK_FREQ	(0x0)
# define I2C_SCL_DELAY	(0x0)
#endif

enum  i2c_int_state {
	I2C_INT_AWAITING,
	I2C_INT_OK,
	I2C_INT_RETRY,
	I2C_INT_ERROR
};

static int i2c_ns9xxx_read( struct i2c_adapter* i2c_adap, int count );
static int i2c_ns9xxx_write( struct i2c_adapter* i2c_adap,const char* buf,int count );
static int i2c_ns9xxx_send_cmd( struct i2c_adapter* i2c_adap, unsigned int cmd );
static void i2c_ns9xxx_if_reset_enable( void );

DECLARE_WAIT_QUEUE_HEAD( wait_q );

/* no multiple instances possible, don't use struct for local variables */
static spinlock_t i2c_ns_lock = SPIN_LOCK_UNLOCKED;/* locking for i2c_ns_buf etc. */
static char* i2c_ns_buf = NULL;	/* where to store the incoming character */
static enum i2c_int_state i2c_state;
static int i2c_clk_freq = I2C_CLK_FREQ;
static unsigned int scl_delay = I2C_SCL_DELAY;

/*
 * Main i2c transfer function, delivers the i2c messages into the bus.
 */
static int i2c_ns9xxx_master_xfer( struct i2c_adapter* i2c_adap,
			       struct i2c_msg msgs[], int num )
{
	int i = 0;
	int ret = 0;
	int len;
	char* buf;
	unsigned long flags;
	int retry = 100;		/* allow only some retries */
	u32 i2c_prev_cfg;

	i2c_state = I2C_INT_OK;
	while( i < num ) {
		if( i2c_state == I2C_INT_RETRY ) {
			/* resend complete message, stop previous access first */
			/* i is already set to 0 due to while condition */
			ret=i2c_ns9xxx_send_cmd(i2c_adap,NS_I2C_DATA_TX_CMD_M_STOP);
			if( ret || !(retry--) ) {
				return -EIO;
			}
		}

		len = msgs[ i ].len;
		buf = msgs[ i ].buf;

		spin_lock_irqsave( &i2c_ns_lock, flags );
		/* being modified by int handler */
		i2c_ns_buf = buf;
		spin_unlock_irqrestore( &i2c_ns_lock, flags );

		if( !(msgs[ i ].flags & I2C_M_NOSTART ) ) {
			unsigned int cmd;

			/* set device address */
			I2C_MSTRAD = ((msgs[i].addr << NS_I2C_MASTER_MDA_SH) & NS_I2C_MASTER_MDA_MA) | ((msgs[i].flags&I2C_M_TEN)?NS_I2C_MASTER_MAM:0);

			if( msgs[ i ].flags & I2C_M_RD )
				cmd = NS_I2C_DATA_TX_CMD_M_READ;
			else {
				/* we need to send first byte with the command */
				cmd = NS_I2C_DATA_TX_CMD_M_WRITE |
					NS_I2C_DATA_TX_VAL |
					*buf;
				len--;
				buf++;
			}

			/* send start, device address and read/write command */
			ret = i2c_ns9xxx_send_cmd( i2c_adap, cmd );
			if( ret ) {
				if( i2c_state == I2C_INT_RETRY ) {
					/* try the message again */
					i = 0;
					continue;
				}
				break;
			}
		}
		if( len > 0 ) {
			/* either read mode or send more than one byte */
			if( msgs[ i ].flags & I2C_M_RD )
				ret = i2c_ns9xxx_read( i2c_adap, len );
			else
				ret = i2c_ns9xxx_write( i2c_adap, buf, len );

			if( ret ) {
				if( i2c_state == I2C_INT_RETRY ) {
					/* try the message again */
					i = 0;
					continue;
				}
				break;
			}
		}

		i++;
	}

	/* message block transmitted, send stop */
	if( i2c_ns9xxx_send_cmd( i2c_adap, NS_I2C_DATA_TX_CMD_M_STOP ) ) {
		/* something locked up, maybe devices thinks lost clock. */
		if( i2c_ns9xxx_send_cmd( i2c_adap, NS_I2C_DATA_TX_CMD_M_NOP ) ||
		    i2c_ns9xxx_send_cmd( i2c_adap, NS_I2C_DATA_TX_CMD_M_STOP ) ) {
			/* maybe ns9xxx I2C state machine got stuck, reset it... */
			i2c_prev_cfg = I2C_CFG;
			printk( KERN_ERR DRIVER_NAME ": resetting i2c interface\n" );
			i2c_ns9xxx_if_reset_enable();
			I2C_CFG = i2c_prev_cfg;
		}
	}

	i2c_ns_buf = NULL;	/* maybe we get spurios interrupts */

	return ret ? 0 : i;
}

static u32 i2c_ns9xxx_master_func( struct i2c_adapter* adap )
{
	return I2C_FUNC_I2C | I2C_FUNC_10BIT_ADDR | I2C_FUNC_SMBUS_QUICK | I2C_FUNC_SMBUS_BYTE |
	    I2C_FUNC_SMBUS_BYTE_DATA | I2C_FUNC_SMBUS_WORD_DATA ;
}

/***********************************************************************
 * @Function: i2c_ns9xxx_irq_handler
 * @Return: IRQ_HANDLED and i2c_state = ERROR or OK
 * @Descr: when data read, stores it in buffer pointed to by i2c_ns_buf. Buffer
 *         is not incremented.
 ***********************************************************************/

static irqreturn_t i2c_ns9xxx_irq_handler(int irq, void* dev_id,struct pt_regs* regs)
{
	int status;

	/* read and acknowledge (by read) */
	status = I2C_RX;

	switch( status & NS_I2C_DATA_RX_IRQCD_MA ) {
	    case NS_I2C_DATA_RX_IRQCD_M_RX_DATA:
		spin_lock( &i2c_ns_lock );
		if( i2c_ns_buf != NULL )
			/* be paranoid for spurios interrupts? */
			*i2c_ns_buf = status & NS_I2C_DATA_RX_DATA_MA;
		spin_unlock( &i2c_ns_lock );
		/* no break, will be done by M_CMD_ACK */

	    case NS_I2C_DATA_RX_IRQCD_M_CMD_ACK:
		i2c_state = I2C_INT_OK;
		break;

	    case NS_I2C_DATA_RX_IRQCD_M_TX_DATA:
		i2c_state = I2C_INT_OK;
		break;

	    case NS_I2C_DATA_RX_IRQCD_M_NO_ACK:
		/* maybe device didn't listen closely, try again */
		i2c_state = I2C_INT_RETRY;
		break;

	    default:
		i2c_state = I2C_INT_ERROR;
                printk( KERN_ERR "%s: Interrupt failed with 0x%x\n",
                        __func__, status );
	}

	wake_up_interruptible( &wait_q );

	return IRQ_HANDLED;
}

/***********************************************************************
 * @Function: i2c_ns9xxx_read
 * @Return: <0 in case of failure otherwise 0
 * @Descr: Must be called after an M_READ has been executed and the
 *         interrupt has happened. Byte has already been stored by the
 *         irq handler in i2c_ns_buf but not incremented. Don't do read
 *         command in i2c_ns9xxx_irq_handler so that we still use
 *         i2c_adap->timeout.
 ***********************************************************************/

static int i2c_ns9xxx_read( struct i2c_adapter* i2c_adap, int count )
{
	int ret = 0;
	unsigned long flags;

	/* first character has already been stored in buffer by interrupt
	 * handler */
	while( count-- > 1 ) {
		spin_lock_irqsave( &i2c_ns_lock, flags );
		i2c_ns_buf++;
		spin_unlock_irqrestore( &i2c_ns_lock, flags );

		ret = i2c_ns9xxx_send_cmd( i2c_adap, NS_I2C_DATA_TX_CMD_M_NOP );
		if( ret )
			return ret;
	}

	return 0;
}

/***********************************************************************
 * @Function: i2c_ns9xxx_write
 * @Return: <0 in case of failure otherwise 0
 * @Descr: Must be called after an M_WRITE has been executed and the
 *         first character has been sent already with M_WRITE.
 ***********************************************************************/

static int i2c_ns9xxx_write( struct i2c_adapter* i2c_adap, const char* buf,int count)
{
	int ret = 0;

	while( count-- ) {
		ret = i2c_ns9xxx_send_cmd( i2c_adap,
				       (*buf & NS_I2C_DATA_TX_DATA_MA ) |
				       NS_I2C_DATA_TX_CMD_M_NOP |
				       NS_I2C_DATA_TX_VAL );
		if( ret )
			return ret;

		buf++;
	}

	return ret;
}

/***********************************************************************
 * @Function: i2c_ns9xxx_send_cmd
 * @Return: -ETIMEDOUT if no interrupt happened in i2c_adap->timeout or -EIO on
 *          an I2C failure
 * @Descr: sends command and waits signal safe for I2C interrupt acknowleding
 *         command or timeout.
 ***********************************************************************/

static int i2c_ns9xxx_send_cmd( struct i2c_adapter* i2c_adap, unsigned int cmd )
{
	int ret = 0;

	/* now wait. Don't use semaphores as these don't support timeouts */
	i2c_state = I2C_INT_AWAITING;
	do {
		I2C_TX = cmd;
		if( !wait_event_interruptible_timeout(
			    wait_q,
			    i2c_state != I2C_INT_AWAITING,
			    i2c_adap->timeout ))
		{
			/* not interrupted */
			ret = -ETIMEDOUT;
			break;
		}
	} while( i2c_state == I2C_INT_AWAITING );

	if( i2c_state != I2C_INT_OK )
		ret = -EIO;

	return ret;
}

static void i2c_ns9xxx_if_reset_enable( void )
{
	if (processor_is_ns9360()) {
		/* Enable the i2c interface*/
		BBUS_MRST |= NS_BBUS_MASTER_RESET_I2C;
		BBUS_MRST &= ~NS_BBUS_MASTER_RESET_I2C;
	}
}

static int __init i2c_ns9xxx_gpio_init( void )
{
	int ret;

	if (processor_is_ns9360()) {
		if ((ret = ns9xxx_request_gpio( 70, DRIVER_NAME )) != 0)
			return ret;
		if ((ret = ns9xxx_request_gpio( 71, DRIVER_NAME )) != 0)
			goto i2c_err_1;
		if ((ret = ns9xxx_gpio_cfgpin(70, GPIO_CFG_FUNC_2)) != 0)
			goto i2c_err_2;
		if ((ret = ns9xxx_gpio_cfgpin(71, GPIO_CFG_FUNC_2)) != 0)
			goto i2c_err_2;

		return 0;
i2c_err_2:
		ns9xxx_release_gpio(71);
i2c_err_1:
		ns9xxx_release_gpio(70);
		return ret;

	} else if (processor_is_ns9750()) {
		/* Nothing to do, NS9750 has specific pins for i2c interface */
		return 0;
	}

	return -ENODEV;

}

static void __init i2c_ns9xxx_gpio_release( void )
{
	if (processor_is_ns9360()) {
		ns9xxx_release_gpio(71);
		ns9xxx_release_gpio(70);
	}
}

static int i2c_ns9xxx_set_clockref( int clock )
{
	unsigned int clref;

	switch (clock) {
		case 100000:	/* normal speed */
			clref = (((ns9xxx_cpuclock()/4)/clock) - 4 - scl_delay) / 2;
			break;

		case 400000:	/* fast speed */
			clref = ((((ns9xxx_cpuclock()/4)/clock) - 4 - scl_delay) * 2) / 3;
			break;
		default:
			return -EINVAL;
	}

	/* Check that value is in range */
	if ((clref & NS_I2C_CFG_CLREF_MA) != clref) {
		printk( KERN_ERR DRIVER_NAME ": unable to set clref = 0x%x\n", clref );
		return -EINVAL;
	}

	I2C_CFG = (I2C_CFG & ~NS_I2C_CFG_CLREF_MA) | (clref & NS_I2C_CFG_CLREF_MA);

	return 0;
}

static struct i2c_algorithm i2c_ns9xxx_algo = {
	.master_xfer	= i2c_ns9xxx_master_xfer,
	.functionality	= i2c_ns9xxx_master_func
};

static struct i2c_adapter i2c_ns_adapter = {
	.owner			= THIS_MODULE,
	.name			= DRIVER_NAME,
	.id			= I2C_HW_NS9750,
	.algo			= &i2c_ns9xxx_algo,
	.timeout		= HZ,
	.retries		= 1
};

static int __init i2c_ns9xxx_init( void )
{
	int ret;

	printk( KERN_INFO DRIVER_NAME
		": I2C NS9XXX driver $Revision: 1.6 $ initializing\n");

	i2c_ns9xxx_if_reset_enable();

	if ((ret = i2c_ns9xxx_gpio_init()) != 0) {
		printk( KERN_ERR DRIVER_NAME ": unable to init GPIOs for i2c interface (%d)\n", ret );
		return ret;
	}

	if( request_mem_region( I2C_BASE_PA, I2C_REG_LEN, DRIVER_NAME ) == NULL ) {
		printk( KERN_ERR DRIVER_NAME ": memory not available\n" );
		ret = -EBUSY;
		goto error;
	}

	if( request_irq( IRQ_I2C, i2c_ns9xxx_irq_handler, 0, DRIVER_NAME, NULL ) ) {
		printk( KERN_ERR DRIVER_NAME ": irq %i not available\n", IRQ_I2C );
		ret = -EBUSY;
		goto error_irq;
	}

	/* With multiple devices connected to the bus, seems to be necessary to increase
	   the value of the spike filter width */
	I2C_CFG = (0xf << NS_I2C_CFG_SFW_SH) & NS_I2C_CFG_SFW_MA;

	if ((ret = i2c_ns9xxx_set_clockref( i2c_clk_freq )) != 0 ) {
		printk( KERN_ERR DRIVER_NAME ": unable to set i2c clock to %i\n", i2c_clk_freq );
		goto error_clk;
	}

	i2c_add_adapter(&i2c_ns_adapter);

	return 0;

error_clk:
	free_irq(IRQ_I2C, NULL);
error_irq:
	release_mem_region(I2C_BASE_PA, I2C_REG_LEN);
error:
	i2c_ns9xxx_gpio_release();
	return ret;
}

static void i2c_ns9xxx_exit( void )
{
	i2c_del_adapter(&i2c_ns_adapter);
	free_irq(IRQ_I2C, NULL);
	release_mem_region(I2C_BASE_PA, I2C_REG_LEN);
}

subsys_initcall( i2c_ns9xxx_init );
module_exit( i2c_ns9xxx_exit );

MODULE_AUTHOR("Digi International Inc");
MODULE_DESCRIPTION( "NS9XXX I2C Driver");
MODULE_LICENSE("GPL");

