/*
 * drivers/i2c/chips/pca9554.c
 *
 * 8 bit I2C GPIO port, with interrupt
 * Derived from pca9539.c by Ben Gardner.
 * 
 * Copyright (C) 2006 by Digi International Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/hwmon-sysfs.h>

/* Addresses to scan */
static unsigned short normal_i2c[] = { 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
				       0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
				       I2C_CLIENT_END };

/* Insmod parameters */
I2C_CLIENT_INSMOD_2(pca9554, pca9554a);

enum pca9554_cmd {
	PCA9554_INPUT		= 0,
	PCA9554_OUTPUT		= 1,
	PCA9554_POLINV		= 2,
	PCA9554_CONFIG		= 3,
};

static int pca9554_attach_adapter(struct i2c_adapter *adapter);
static int pca9554_detect(struct i2c_adapter *adapter, int address, int kind);
static int pca9554_detach_client(struct i2c_client *client);

/* This is the driver that will be inserted */
static struct i2c_driver pca9554_driver = {
	.driver = {
		.name	= "pca9554(a)",
	},
	.attach_adapter	= pca9554_attach_adapter,
	.detach_client	= pca9554_detach_client,
};

struct pca9554_data {
	struct i2c_client client;
};

/* following are the sysfs callback functions */
static ssize_t pca9554_show(struct device *dev, struct device_attribute *attr,
			    char *buf)
{
	struct sensor_device_attribute *psa = to_sensor_dev_attr(attr);
	struct i2c_client *client = to_i2c_client(dev);
	return sprintf(buf, "%d\n", i2c_smbus_read_byte_data(client,
							     psa->index));
}

static ssize_t pca9554_store(struct device *dev, struct device_attribute *attr,
			     const char *buf, size_t count)
{
	struct sensor_device_attribute *psa = to_sensor_dev_attr(attr);
	struct i2c_client *client = to_i2c_client(dev);
	unsigned long val = simple_strtoul(buf, NULL, 0);
	if (val > 0xff)
		return -EINVAL;
	i2c_smbus_write_byte_data(client, psa->index, val);
	return count;
}

/* Define the device attributes */
#define PCA9554_ENTRY_RO(name, cmd_idx) \
	static SENSOR_DEVICE_ATTR(name, S_IRUGO, pca9554_show, NULL, cmd_idx)

#define PCA9554_ENTRY_RW(name, cmd_idx) \
	static SENSOR_DEVICE_ATTR(name, S_IRUGO | S_IWUSR, pca9554_show, \
				  pca9554_store, cmd_idx)

PCA9554_ENTRY_RO(input, PCA9554_INPUT);
PCA9554_ENTRY_RW(output, PCA9554_OUTPUT);
PCA9554_ENTRY_RW(polinv, PCA9554_POLINV);
PCA9554_ENTRY_RW(config, PCA9554_CONFIG);

static struct attribute *pca9554_attributes[] = {
	&sensor_dev_attr_input.dev_attr.attr,
	&sensor_dev_attr_output.dev_attr.attr,
	&sensor_dev_attr_polinv.dev_attr.attr,
	&sensor_dev_attr_config.dev_attr.attr,
	NULL
};

static struct attribute_group pca9554_defattr_group = {
	.attrs = pca9554_attributes,
};

static int pca9554_attach_adapter(struct i2c_adapter *adapter)
{
	return i2c_probe(adapter, &addr_data, pca9554_detect);
}

/* This function is called by i2c_probe */
static int pca9554_detect(struct i2c_adapter *adapter, int address, int kind)
{
	struct i2c_client *new_client;
	struct pca9554_data *data;
	int err = 0;

	if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE_DATA))
		goto exit;

	/* OK. For now, we presume we have a valid client. We now create the
	   client structure, even though we cannot fill it completely yet. */
	if (!(data = kzalloc(sizeof(struct pca9554_data), GFP_KERNEL))) {
		err = -ENOMEM;
		goto exit;
	}

	new_client = &data->client;
	i2c_set_clientdata(new_client, data);
	new_client->addr = address;
	new_client->adapter = adapter;
	new_client->driver = &pca9554_driver;
	new_client->flags = 0;

	/* Detection: the pca9554 only has 4 registers (0-3).
	   A read of 3 should succeed, but a read of 4 should fail. */
	if ((i2c_smbus_read_byte_data(new_client, 3) < 0) ||
	    (i2c_smbus_read_byte_data(new_client, 4) >= 0))
		goto exit_kfree;

	strlcpy(new_client->name, "pca9554(a)", I2C_NAME_SIZE);

	/* Tell the I2C layer a new client has arrived */
	if ((err = i2c_attach_client(new_client)))
		goto exit_kfree;

	/* Register sysfs hooks (don't care about failure) */
	sysfs_create_group(&new_client->dev.kobj, &pca9554_defattr_group);

	return 0;

exit_kfree:
	kfree(data);
exit:
	return err;
}

static int pca9554_detach_client(struct i2c_client *client)
{
	int err;

	if ((err = i2c_detach_client(client)))
		return err;

	kfree(i2c_get_clientdata(client));
	return 0;
}

static int __init pca9554_init(void)
{
	return i2c_add_driver(&pca9554_driver);
}

static void __exit pca9554_exit(void)
{
	i2c_del_driver(&pca9554_driver);
}

MODULE_AUTHOR("Digi International Inc.");
MODULE_DESCRIPTION("PCA9554(A) driver");
MODULE_LICENSE("GPL");

module_init(pca9554_init);
module_exit(pca9554_exit);

