/************************************************************************
 *
 * Copyright (C) 2005 by FS Forth-Systeme GmbH && Sistemas Embebidos S.A.
 * All rights reserved.
 *
 * $Id: a9m24x0_dev-ide.c,v 1.5 2005-10-26 15:43:30 jjaeger Exp $
 * @Author: Pedro Perez de Heredia
 * @Descr: IDE driver for the CompactFlash interface for the A9M24X0
 *         module. Based on bast-ide.c
 * @References: [1] Compact Flash Interface for MODARM9 specification.
 *                  Version 1.0.
 *
 * ChangeLog
 *
 * 2005-08-08: Joachim Jaeger
 * 	- updated to the 2.6.12.4 kernel
 *
 * 2005-10-26: Joachim Jaeger
 * 	- IRQ changed from low level to falling edge detection,
 *	  because of lost interrupts at several CF cards
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ************************************************************************/

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/ide.h>
#include <linux/init.h>
#include <linux/delay.h>

#include <asm/mach-types.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/arch/map.h>
#include <asm/arch/smdk2410.h>
#include <asm/arch/regs-mem.h>

#define	A9M24X0_MAX_IDE_HWIF		1
#define CF_SHIFT_ADDR			1
#define	A9M2410_DEV_KIT_IDE_IRQ		IRQ_EINT14
#define	CF_CPLD_BASE_ADDR		S3C2410_CS4

// CPLD internal addresses (Base address is nGCS4)
#define	CF_CPLD_IO_8_BIT_SPACE		0x000
#define	CF_CPLD_IO_16_BIT_SPACE		0x800
#define	CF_CPLD_CONF_REG_OF		0x1000
#define	CF_CPLD_LOGIC_REV_OF		0x1002
#define	CF_CPLD_LOGIC_MEM_LEN		0x1fff
#define	CF_CPLD_HWIF_OFFSET		0x2000

// CF configuration register flags
#define CF_CD_MASK			0x0001
#define CF_IRQ_MASK			0x0002
#define CF_RST_MASK			0x0004
#define CF_REG_MASK			0x0008
#define CF_MODE_MASK			0x0010
#define CF_16B_MASK			0x0020

// Common Memory/IO internal registers for Storage CF+ cards
#define CF_DATA_OFFSET			(0x00 << CF_SHIFT_ADDR)	// Data
#define CF_ERROR_OFFSET			(0x01 << CF_SHIFT_ADDR)	// Error
#define CF_NSECTOR_OFFSET		(0x02 << CF_SHIFT_ADDR)	// Number of sectors
#define CF_SECTOR_OFFSET		(0x03 << CF_SHIFT_ADDR)	// Starting sector
#define CF_LCYL_OFFSET			(0x04 << CF_SHIFT_ADDR)	// Low Cylinder
#define CF_HCYL_OFFSET			(0x05 << CF_SHIFT_ADDR)	// High Cylinder
#define CF_SELECT_OFFSET		(0x06 << CF_SHIFT_ADDR)	// Select
#define CF_STATUS_OFFSET		(0x07 << CF_SHIFT_ADDR)	// Status
#define CF_CTRL_OFFSET			(0x0e << CF_SHIFT_ADDR)	// Control
#define CF_DATA_DUP_OFFSET		(0x08 << CF_SHIFT_ADDR)	// Duplicate Data Register

#define CF_CONFIG_OPTION_REG_OF	0x200	// Compact Flash Storage card Configuration
					// Option register


/* list of registered interfaces */
static ide_hwif_t *hwifs[A9M24X0_MAX_IDE_HWIF];

static int __init
a9m24x0_ide_register( unsigned int base, unsigned int aux, int irq,
		 ide_hwif_t **hwif)
{
	hw_regs_t hw;

	memset( &hw, 0, sizeof(hw) );

	// Configure addresses of IO registers
	// Use duplicate data register instead of data register
	hw.io_ports[IDE_ERROR_OFFSET] = (unsigned long)( base + CF_ERROR_OFFSET );		// Error
	hw.io_ports[IDE_NSECTOR_OFFSET] = (unsigned long)( base + CF_NSECTOR_OFFSET );		// Sector Count
	hw.io_ports[IDE_SECTOR_OFFSET] = (unsigned long)( base + CF_SECTOR_OFFSET );		// Sector No
	hw.io_ports[IDE_LCYL_OFFSET] = (unsigned long)( base + CF_LCYL_OFFSET );		// Low Cylinder
	hw.io_ports[IDE_HCYL_OFFSET] = (unsigned long)( base + CF_HCYL_OFFSET );		// High Cylinder
	hw.io_ports[IDE_SELECT_OFFSET] = (unsigned long)( base + CF_SELECT_OFFSET );		// Select
	hw.io_ports[IDE_STATUS_OFFSET] = (unsigned long)( base + CF_STATUS_OFFSET );		// Status
	hw.io_ports[IDE_CONTROL_OFFSET] = (unsigned long)( base + CF_CTRL_OFFSET );		// Control

	// For the data register, access over the 16bit IO space
	hw.io_ports[IDE_DATA_OFFSET] = (unsigned long)( base +
	                                                CF_DATA_DUP_OFFSET +
	                                                CF_CPLD_IO_16_BIT_SPACE );		// Data

	hw.irq = irq;

	ide_register_hw( &hw, hwif );

	return 0;
}

static int __init a9m24x0_ide_init(void)
{
	unsigned long ulReg;
	unsigned long ulBaseAddr;
	unsigned short ucCtrlReg;
	unsigned char ucLogicRev;
	int nCounter;

	if( !(machine_is_a9m2410() || machine_is_a9m2440()) )
		return 0;


	// Perform initial hardware configuration

	// Configure the nGCS4 to 16 bit and wait enabled
	ulReg = readl( S3C24XX_VA_MEMCTRL );
        writel( (ulReg & 0xfff0ffff) | S3C2410_BWSCON_DW4_16 | S3C2410_BWSCON_WS4,
	        S3C24XX_VA_MEMCTRL );

	// Remap physical to virtual mem
	ulBaseAddr = (unsigned long)ioremap( CF_CPLD_BASE_ADDR, CF_CPLD_LOGIC_MEM_LEN );

	ucLogicRev = readb( ulBaseAddr + CF_CPLD_LOGIC_REV_OF );
	printk( "A9M24X0_dev-ide Logic Revison: %d.%d\n", ucLogicRev >> 4, ucLogicRev & 0x0f );

	// Configure IRQ line
	set_irq_type( A9M2410_DEV_KIT_IDE_IRQ, IRQT_FALLING );

	for( nCounter = 0; nCounter < A9M24X0_MAX_IDE_HWIF; nCounter++ )
	{
	    // Detect if a CF is inserted on the socket
	    ucCtrlReg = readb( ulBaseAddr + CF_CPLD_CONF_REG_OF + CF_CPLD_HWIF_OFFSET * nCounter );

	    if( ( ucCtrlReg & CF_CD_MASK ) != CF_CD_MASK )
	    {
	        printk( "CF detected\n" );

	        // Reset CF
	        writeb( CF_RST_MASK, ulBaseAddr + CF_CPLD_CONF_REG_OF + CF_CPLD_HWIF_OFFSET * nCounter );
	        mdelay(10);
	        writeb( 0x00, ulBaseAddr + CF_CPLD_CONF_REG_OF + CF_CPLD_HWIF_OFFSET * nCounter );
	        mdelay(10); // @TODO review following delay...

	        // Set CF in IO mode and activate CF IRQ to work in level mode
	        writeb( 0x41, ulBaseAddr + CF_CONFIG_OPTION_REG_OF + CF_CPLD_HWIF_OFFSET * nCounter );

	        //Check if the value writen can be read (some slow CF doesn't work...)
	        if( readb( ulBaseAddr + CF_CONFIG_OPTION_REG_OF + CF_CPLD_HWIF_OFFSET * nCounter ) != 0x41 )
	        {
	            printk( "A9M24X0_dev-ide: unable to access to Configuration Option Register\n" );
	            continue;
	        }

	        // Configure CPLD to operate in IO mode
	        writeb( CF_MODE_MASK | CF_16B_MASK, ulBaseAddr + CF_CPLD_CONF_REG_OF + CF_CPLD_HWIF_OFFSET * nCounter );

	        a9m24x0_ide_register( ulBaseAddr + CF_CPLD_HWIF_OFFSET * nCounter + CF_CPLD_IO_8_BIT_SPACE,
	                              ulBaseAddr + CF_CPLD_HWIF_OFFSET * nCounter + CF_CPLD_IO_8_BIT_SPACE,
	                              A9M2410_DEV_KIT_IDE_IRQ, &hwifs[nCounter] );
	    }
	}

	return 0;
}

module_init( a9m24x0_ide_init );

MODULE_AUTHOR( "Pedro Perez de Heredia <pperez@embebidos.com>" );
MODULE_LICENSE( "GPL" );
MODULE_DESCRIPTION( "A9M24X0_DEV IDE driver");
