/***********************************************************************
 *
 * Copyright (C) 2004 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * $Id: ns9750_eth.c,v 1.13 2006-10-09 20:05:01 pperez Exp $
 * @Author: Markus Pietrek
 * @Descr: Ethernet driver for the NS9750.
 * @References: [1] NS9750 Hardware Reference, December 2003
 *              [2] Intel LXT971 Datasheet #249414 Rev. 02
 *              [3] NS7520 Linux Ethernet Driver
 *              [4] u-boot-1.0.2/drivers/ns9750_eth.c
 *              [5] drivers/net/pci_skeleton.c
 *              [6] drivers/net/isa_skeleton.c
 *              [7] appnote_ethernetrx_9750.pdf
 * @TODO:
 *    o add media change interrupts, cable looses etc.
 *    o split sizes for RX DMA  descriptors
 *    o add locking
 *    o check retransmit bug
 *    o in case of a timeout we should retransmit the frame
 *    o implement STAT overflow handling in interrupt handler
 *    o add scatter-gather (NETIF_F_SG)
 *      http://www.spinics.net/lists/linux-net/msg08720.html
 *    o fix UDP checksum failure (happens with nfsroot/portmap) with
 *      NETIF_F_HW_CSUM enabled in features
 *      http://tonnikala.net/2003/41/linux260test7.php
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***********************************************************************/
/***********************************************************************
 *  @History:
 *	2006/08/08 : CCW9C added (Bernd Westermann)
 *	2006/05/11 : Ethernet Receive Data FIFO Overflow fixed [7]
 *	2005/09/14 : CC9C added
 *	2005/06/02 : multicast, promiscuous and all multicast mode added (Bernd Westermann)
 *	2005/03/04 : NS9360 added
 *     2004/09/01 : BugID 314: fixed caching problems
 *     2004/08/17 : removed addr_
 ***********************************************************************/
#include <linux/config.h>	/* system configuration */

#include <asm/arch-ns9750/ns9750_eth.h>


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/ioport.h>	/* request_mem_region */
#include <linux/netdevice.h>	/* net_device */
#include <linux/etherdevice.h>
#include <linux/delay.h>
#include <linux/ethtool.h>
#include <linux/mii.h>
#include <linux/crc32.h>
#include <linux/sockios.h>

#include <asm/io.h>

#include <asm/irq.h>		/* enable_interrupt */

#include <asm/arch/system.h> /* ns9xxx_ahbclock */

#ifdef CONFIG_ARCH_NS9XXX
#include <asm/arch/processor.h>
#endif

#ifndef rmwl
#define rmwl(addr,op)   writel(readl(addr) op, (addr) )
#endif

#ifndef IRQ_ETH_RX
#define IRQ_ETH_RX IRQ_ETHRX
#endif
#ifndef IRQ_ETH_TX
#define IRQ_ETH_TX IRQ_ETHTX
#endif

#undef DEBUG

#ifdef	DEBUG
# define DEBUG_INIT		0x0001
# define DEBUG_MINOR		0x0002
# define DEBUG_RX		0x0004
# define DEBUG_TX		0x0008
# define DEBUG_INT		0x0010
# define DEBUG_POLL		0x0020
# define DEBUG_LINK		0x0040
# define DEBUG_MII		0x0100
# define DEBUG_MII_LOW		0x0200
# define DEBUG_MEM		0x0400
# define DEBUG_ERROR		0x4000
# define DEBUG_ERROR_CRIT	0x8000

/*static int debug = DEBUG_ERROR_CRIT; */
static int debug_level = DEBUG_ERROR_CRIT;

# define DEBUG_ARGS0( FLG, a0 ) if( ( debug_level & (FLG) ) == (FLG) ) \
		printk( KERN_INFO "%s: " a0,  __FUNCTION__ )
# define DEBUG_ARGS1( FLG, a0, a1 ) if( ( debug_level & (FLG) ) == (FLG)) \
		printk( KERN_INFO "%s: " a0, __FUNCTION__, (a1) )
# define DEBUG_ARGS2( FLG, a0, a1, a2 ) if( (debug_level & (FLG)) ==(FLG))\
		printk( KERN_INFO "%s: " a0, __FUNCTION__, (a1), (a2) )
# define DEBUG_ARGS3( FLG, a0, a1, a2, a3 ) if((debug_level &(FLG))==(FLG))\
		printk(KERN_INFO "%s: "a0, __FUNCTION__,(a1),(a2),(a3))
# define DEBUG_ARGS4( FLG, a0, a1, a2, a3, a4 ) if((debug_level &(FLG))==(FLG))\
		printk(KERN_INFO "%s: "a0, __FUNCTION__,(a1),(a2),(a3),(a4))
# define DEBUG_ARGS6( FLG, a0, a1, a2, a3, a4, a5, a6 ) if((debug_level &(FLG))==(FLG))\
		printk(KERN_INFO "%s: "a0, __FUNCTION__,(a1),(a2),(a3),(a4),(a5),(a6))
# define DEBUG_FN( FLG ) if( (debug_level & (FLG)) == (FLG) ) \
		printk( KERN_INFO "\r%s:line %d\n",__FUNCTION__,__LINE__)
# define ASSERT( expr, func ) if( !( expr ) ) { \
        	printk( KERN_ERR "Assertion failed! %s:line %d %s\n", \
			__FUNCTION__,__LINE__,#expr);		      \
        	func }
#else /* DEBUG */
// # define printk(...) do {} while( 0 )
# define DEBUG_ARGS0( FLG, a0 ) do {} while( 0 )
# define DEBUG_ARGS1( FLG, a0, a1 ) do {} while( 0 )
# define DEBUG_ARGS2( FLG, a0, a1, a2 ) do {} while( 0 )
# define DEBUG_ARGS3( FLG, a0, a1, a2, a3 ) do {} while( 0 )
# define DEBUG_ARGS4( FLG, a0, a1, a2, a3, a4 ) do {} while( 0 )
# define DEBUG_ARGS6( FLG, a0, a1, a2, a3, a4, a5, a6 ) do {} while( 0 )
# define DEBUG_FN( n ) do {} while( 0 )
# define ASSERT( expr, func ) do {} while( 0 )
#endif /* DEBUG */

#define NS_TIMEOUT		5*HZ /* in jiffies */
#define NS_MII_NEG_DELAY	5*HZ  /* in jiffies */
#define DRIVER_NAME		"ns_eth"

#define NS_TX_DESC		64
#define NS_RX_RINGS		4
#define NS_RX_DESC_RING		60 /* descriptors per ring */
#define NS_RX_DESC		(NS_RX_RINGS*NS_RX_DESC_RING) /* overall desc */
#define NS_ETH_FRAME_LEN	1524 /* NS9750 accepts 1522 byte packets */

#define NS_POLYNOMIAL		  0x4c11db6L // taken from [1] p183
#define NS_LEFT			  0          // left rotation
#define NS_RIGHT		  1          // right rotation

/* @TODO move it to eeprom.h */
#define FS_EEPROM_AUTONEG_MASK		0x7
#define FS_EEPROM_AUTONEG_SPEED_MASK	0x1
#define FS_EEPROM_AUTONEG_SPEED_10	0x0
#define FS_EEPROM_AUTONEG_SPEED_100	0x1
#define FS_EEPROM_AUTONEG_DUPLEX_MASK	0x2
#define FS_EEPROM_AUTONEG_DUPLEX_HALF	0x0
#define FS_EEPROM_AUTONEG_DUPLEX_FULL	0x2
#define FS_EEPROM_AUTONEG_ENABLE_MASK	0x4
#define FS_EEPROM_AUTONEG_DISABLE	0x0
#define FS_EEPROM_AUTONEG_ENABLE	0x4

#define SET_HWADDRESS (SIOCDEVPRIVATE +1)
/* buffer descriptors taken from [1] p.306 */
typedef struct
{
	unsigned int* src;
	unsigned int len;	/* 11 bits */
	unsigned int* dest;	/* unused */
	union {
		unsigned int reg;
		struct {
			unsigned status : 16;
			unsigned res : 12;
			unsigned full : 1;
			unsigned enable : 1;
			unsigned intr : 1;
			unsigned wrap : 1;
		} bits;
	} s;
} rx_buffer_desc_t;

typedef struct
{
	unsigned int* src;
	unsigned int len;	/* 10 bits */
	unsigned int* dest;	/* unused */
	union {
		unsigned int reg; /* only 32bit accesses may done to NS9750
				   * eth engine */
		struct {
			unsigned status : 16;
			unsigned res : 12;
			unsigned full : 1;
			unsigned last : 1;
			unsigned intr : 1;
			unsigned wrap : 1;
		} bits;
	} s;
} tx_buffer_desc_t;

/* ns_eth_private_t contains private data to this module */
typedef struct
{
	struct sk_buff* tx_skb[ NS_TX_DESC ]; /* transmit packet buffers */
	struct sk_buff* rx_skb[ NS_RX_DESC ]; /* receive packet buffers */

	dma_addr_t tx_dma[ NS_TX_DESC ];      /* for uncaching DMA */
	dma_addr_t rx_dma[ NS_RX_DESC ];      /* for uncaching DMA */

	/* hw descriptors. The first NS_RX_DESC_RING are for ring desc A, then
	 * B etc. No multiarray to keep iterators simpler. */
	rx_buffer_desc_t*  rx_desc; /* [NS_RX_DESC] */
	dma_addr_t 	   rx_dma_desc;

	int		tx_dma_next; /* start descr for next DMA transfer */
	int		tx_dma_start; /* start descr for current DMA transfer */
	int		tx_dma_free; /* next free descriptor  */

	unsigned short last_link_status; /* stores the last detected link
					  * information 10/100 half/full auto
					  * 0xff is empty */

	spinlock_t lock;	/* locking for multiple processes */

} ns_eth_private_t;

static int ns_eth_open( struct net_device* dev );
static int ns_eth_close( struct net_device* dev );
static int ns_eth_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd);
static int ns_eth_start_xmit( struct sk_buff* skb, struct net_device* dev );
static irqreturn_t ns_eth_rx_int( int irq, void* dev_id,struct pt_regs* regs );
static irqreturn_t ns_eth_tx_int( int irq, void* dev_id,struct pt_regs* regs );
static inline void ns_eth_rx_process_ring( struct net_device* dev, int ring );
static struct net_device_stats* ns_eth_get_stats( struct net_device* dev );
static void ns_eth_set_multicast_list( struct net_device* dev );
static int ns_eth_set_mac_address(struct net_device *dev, void *p);
static void ns_eth_tx_timeout( struct net_device* dev );
static void ns_irqs_disable( void );
static void ns_irqs_enable( void );
static int ns_eth_reset( struct net_device* dev );
static void ns_eth_enable( struct net_device* dev );
static void ns_eth_update_rx_desc( struct net_device* dev );

static void ns_link_force( struct net_device* dev );
static void ns_link_auto_negotiate(  struct net_device* dev );
static void ns_link_update_egcr( struct net_device* dev );
static void ns_link_print_changed( struct net_device* dev );

/* the PHY stuff */

static char ns_mii_identify_phy( void );
static unsigned short ns_mii_read( unsigned short reg );
static void ns_mii_write( unsigned short reg, unsigned short data );
static unsigned int ns_mii_get_clock_divisor( unsigned int max_mdio_clk );
static unsigned int ns_mii_poll_busy( void );

// the multicast stuff

static void ns_load_mca_hash_table( struct dev_mc_list* pMcList );
static void ns_set_hash_bit( unsigned char* pucHashTable, int nBit );
static int ns_calculate_hash_bit( unsigned char* pucMac );
static unsigned short ns_rotate( unsigned short uiReg,
				     int nDirection,
				     int nBits );

/* the module stuff */
static int __init  ns_init_device( struct net_device* dev );
static int __init  ns_eth_init_module( void );
static void __exit ns_eth_cleanup_module( void );

static struct net_device ns_eth_device = {
	init: 			ns_init_device,
	open: 			ns_eth_open,
	stop: 			ns_eth_close,
	do_ioctl:			ns_eth_ioctl,
	hard_start_xmit: 	ns_eth_start_xmit,
	get_stats: 		ns_eth_get_stats,
	set_multicast_list: 	ns_eth_set_multicast_list,
	set_mac_address:	ns_eth_set_mac_address,
	tx_timeout: 		ns_eth_tx_timeout,

	watchdog_timeo: 	NS_TIMEOUT,
	base_addr: 		NS_ETH_MODULE_BASE,
/*	features: 		NETIF_F_HW_CSUM,*//*@TODO: UDP checksum failure*/
	flags:			IFF_MULTICAST | IFF_BROADCAST
};

static const char* version = "Ethernet: NS9xx0 driver $Revision: 1.13 $\n";

static unsigned int phy_mdio_max_clk = PHY_MDIO_MAX_CLK;
static unsigned char link_mode       = FS_EEPROM_AUTONEG_ENABLE;
static phy_type_t phy_detected       = PHY_NONE;

static unsigned short NS_ETH_PHY_ADDRESS = 0;
struct net_device_stats ns_eth_stats;

static int ns_eth_open( struct net_device* dev )
{
	int ret = -EBUSY;
	ns_eth_private_t* priv;
	int i;

	DEBUG_FN( DEBUG_INIT );

	ASSERT( NULL == 0, return -ENODEV; );	/* for rx_skb allocation */

	/* reset configuration from U-Boot or previous "ifconfig up" */
	writel( 0x0, NS_ETH_EGCR1 );
	writel( 0x0, NS_ETH_EINTREN );
	writel( readl( NS_ETH_EINTR ), NS_ETH_EINTR );

	if( request_mem_region( NS_ETH_MODULE_BASE,
				NS_ETH_TXBD + 0x400 - NS_ETH_MODULE_BASE,
				DRIVER_NAME ) == NULL )
		goto error_region;

	if( request_irq( IRQ_ETH_RX, ns_eth_rx_int, 0, DRIVER_NAME " rx", dev ) )
		goto error_irq_rx;

	if( request_irq( IRQ_ETH_TX, ns_eth_tx_int, 0, DRIVER_NAME " tx", dev ) )
		goto error_irq_tx;
        ns_irqs_disable();

	/* allocate private field */
	priv = dev->priv = kmalloc( sizeof( ns_eth_private_t ), GFP_KERNEL);
	if( priv == NULL )
	{
	    ret = -ENOMEM;
	    goto error_nomem;
	}
	memset( priv, 0, sizeof( ns_eth_private_t ) ); /* inits tx_dma_* etc. */

	spin_lock_init( &priv->lock );

	priv->rx_desc = (rx_buffer_desc_t*)
		dma_alloc_coherent( NULL,
				    sizeof( rx_buffer_desc_t ) * NS_RX_DESC,
				    &priv->rx_dma_desc, GFP_DMA );
	if( priv->rx_desc == NULL ) {
		ret = -ENOMEM;
		goto error_nomem_rx;
	}

	/* preallocate any receive skb buffers for DMA transfer
	   we receive the rx interrupt after packets have been copied! */
	for( i = 0; i < NS_RX_DESC; i++ )
	{
		priv->rx_skb[ i ] = dev_alloc_skb( NS_ETH_FRAME_LEN + 2 );
		if( priv->rx_skb[ i ] == NULL )
		{
			ret = -ENOMEM;
			goto error;
		}
		priv->rx_skb[ i ]->dev = dev;

		/* prepare descriptor */
		priv->rx_desc[ i ].src =
			(unsigned int*) virt_to_bus( priv->rx_skb[ i ]->data );
		priv->rx_desc[ i ].len = NS_ETH_FRAME_LEN;
		/* mark only the last desc in the ring to be wrap */
		priv->rx_desc[ i ].s.bits.wrap=!((i+1)%(NS_RX_DESC_RING))?1:0;
		priv->rx_desc[ i ].s.bits.intr = 0;
		priv->rx_desc[ i ].s.bits.enable = 1;
		priv->rx_desc[ i ].s.bits.full = 0;
	}

	/* all ok, continue with hardware initialization
	   from here we have access to ports */

	if( !ns_eth_reset( dev ) )
	{
	    ret = -ENODEV;
	    goto error;
	}

/*	ns7520_rx_lockup_timer_enable( pDev ); */  /* @TODO */

	/* assign hardware address */

	writel( dev->dev_addr[ 5 ] << 8 |
		dev->dev_addr[ 4 ], NS_ETH_SA1 );
	writel( dev->dev_addr[ 3 ] << 8 |
		dev->dev_addr[ 2 ], NS_ETH_SA2 );
	writel( dev->dev_addr[ 1 ] << 8 |
		dev->dev_addr[ 0 ], NS_ETH_SA3 );

	/* Receive broadcasts, so we can see ARP requests */
	rmwl( NS_ETH_SAFR, | NS_ETH_SAFR_BROAD );

	ns_eth_enable( dev );
        ns_irqs_enable();

	netif_start_queue( dev );

	return 0;

error:
	/* free skb buffers */
	for( i = 0; i < NS_RX_DESC; i++ )
	{
		if( priv->rx_skb[ i ] != NULL ) {
			dev_kfree_skb( priv->rx_skb[ i ] );
		} else
			/* no more buffers allocated */
			break;
	}

	dma_free_coherent( NULL, sizeof( rx_buffer_desc_t ) * NS_RX_DESC,
			   priv->rx_desc, priv->rx_dma_desc );
error_nomem_rx:
	kfree( priv );
error_nomem:
	free_irq( IRQ_ETH_TX, dev );
error_irq_tx:
	free_irq( IRQ_ETH_RX, dev );
error_irq_rx:
	release_mem_region( NS_ETH_MODULE_BASE,
			    NS_ETH_TXBD + 0x400 - NS_ETH_MODULE_BASE );
error_region:
	return ret;
}

static int ns_eth_close( struct net_device* dev )
{
	ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
	int i;

	DEBUG_FN( DEBUG_INIT );

        ns_irqs_disable();

	netif_stop_queue( dev ); /* can't transmit any more */

	/* disable hardware */

	rmwl( NS_ETH_MAC1, & ~NS_ETH_MAC1_RXEN );
	rmwl( NS_ETH_EGCR1, & ~(NS_ETH_EGCR1_ERX |
				NS_ETH_EGCR1_ERXDMA |
				NS_ETH_EGCR1_ETX |
				NS_ETH_EGCR1_ETXDMA ) );
	/* release ressources */
	free_irq( IRQ_ETH_TX, dev );
	free_irq( IRQ_ETH_RX, dev );
	release_mem_region( NS_ETH_MODULE_BASE,
			    NS_ETH_TXBD + 0x400 - NS_ETH_MODULE_BASE );

	/* DMA buffers */
	for( i = 0; i < NS_RX_DESC; i++ )
		dev_kfree_skb( priv->rx_skb[ i ] );

	for( i = 0; i < NS_TX_DESC; i++ )
		if( priv->tx_skb[ i ] != NULL ) {
			/* we won't transmit it any longer */
			dma_unmap_single( NULL, priv->tx_dma[ i ],
					  priv->tx_skb[ i ]->len,DMA_TO_DEVICE);
			dev_kfree_skb( priv->tx_skb[ i ] );
		}

	dma_free_coherent( NULL, sizeof( rx_buffer_desc_t ) * NS_RX_DESC,
			   priv->rx_desc, priv->rx_dma_desc );

	kfree( dev->priv );

	return 0;
}

static int ns_eth_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	switch (cmd)
	{
		case SET_HWADDRESS:
			ns_eth_set_mac_address(dev, ifr->ifr_hwaddr.sa_data);
			break;
		default:
			return -ENOIOCTLCMD;
	}

	return 0;
}

static int ns_eth_start_xmit( struct sk_buff* skb, struct net_device* dev )
{
	ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
	unsigned long flags;
	tx_buffer_desc_t* tx_buffer_desc;

	DEBUG_FN( DEBUG_TX );

        spin_lock_irqsave( &priv->lock, flags );

	/* check that we don't overwrite */
	ASSERT( priv->tx_skb[priv->tx_dma_free] == NULL, );

	priv->tx_skb[ priv->tx_dma_free ] = skb;
	priv->tx_dma[ priv->tx_dma_free ] = dma_map_single( NULL, skb->data,
							    skb->len,
							    DMA_TO_DEVICE );
	/* configure next free buffer descriptor */
	tx_buffer_desc = (tx_buffer_desc_t*) NS_ETH_TXBD;
	tx_buffer_desc += priv->tx_dma_free;

	/* skb->data is 32bit aligned */
	tx_buffer_desc->src = (unsigned int*) priv->tx_dma[ priv->tx_dma_free ];
	tx_buffer_desc->len = (skb->len >= ETH_ZLEN ? skb->len : ETH_ZLEN);

	/* only 32bit accesses allowed.wrap, last, full  */
	tx_buffer_desc->s.reg = 0x00000000;

	if( priv->tx_dma_free == NS_TX_DESC - 1 )
		/* descriptor table full, wait until isr cleared them */
		/* @TODO: we could use the wraparound feature and use the
		 * descriptor table as a ring fifo to tx_start-1 */
		netif_stop_queue( dev );

	if( !priv->tx_dma_free ) {
		/* no packages have been added while doing a transfer, begin at
		 * start of descriptor table */
		priv->tx_dma_next = priv->tx_dma_free + 1;
		tx_buffer_desc->s.reg |= 0xf0000000;

		/* enable DMA engine to transmit descriptor */

		writel( 0, NS_ETH_TXPTR );
		writel( 0, NS_ETH_TXRPTR );
		rmwl( NS_ETH_EGCR1, | NS_ETH_EGCR1_ETXDMA );
		rmwl( NS_ETH_EGCR2, & ~NS_ETH_EGCR2_TCLER );
		rmwl( NS_ETH_EGCR2, | NS_ETH_EGCR2_TCLER );
	}

	dev->trans_start = jiffies;

	priv->tx_dma_free++;

        spin_unlock_irqrestore( &priv->lock, flags );

	return 0;

}

static void ns_eth_recover_rxovl_stall( struct net_device* dev )
{
        ns_eth_private_t* priv = dev->priv;
	int i;
	
	rmwl( NS_ETH_EGCR1, & ~(NS_ETH_EGCR1_ERX | NS_ETH_EGCR1_ERXDMA ) );
	rmwl( NS_ETH_MAC1, & ~NS_ETH_MAC1_RXEN );
	rmwl( NS_ETH_EINTR, | NS_ETH_EINTR_RXOVL_DATA );
	
	for( i = 0; i < NS_RX_DESC; i++ )
	{		
		priv->rx_desc[ i ].len = NS_ETH_FRAME_LEN;
		/* mark only the last desc in the ring to be wrap */
		priv->rx_desc[ i ].s.bits.wrap=!((i+1)%(NS_RX_DESC_RING))?1:0;
		priv->rx_desc[ i ].s.bits.intr = 0;
		priv->rx_desc[ i ].s.bits.enable = 1;
		priv->rx_desc[ i ].s.bits.full = 0;
	}
	
	for( i = 0; i < NS_RX_RINGS; i++ )
		writel((unsigned int) (priv->rx_dma_desc + i*NS_RX_DESC_RING),
			NS_ETH_RXAPTR + i * 4 );
	
	rmwl( NS_ETH_EGSR, | NS_ETH_EGSR_RXINIT );
	rmwl( NS_ETH_EGCR1, | NS_ETH_EGCR1_ERX );
	
	ns_eth_update_rx_desc( dev );
	
	rmwl( NS_ETH_MAC1, | NS_ETH_MAC1_RXEN );
	
}

static irqreturn_t ns_eth_rx_int( int irq, void* dev_id,struct pt_regs* regs )
{
	struct net_device* dev = (struct net_device*) dev_id;
	ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
	unsigned int status;
	int ring;
	int mask;

	DEBUG_FN( DEBUG_RX | DEBUG_INT );

        spin_lock( &priv->lock );

	/* acknowledge interrupt */
	status = readl( NS_ETH_EINTR ) & NS_ETH_EINTR_RX_MA;
	writel( status, NS_ETH_EINTR );

	if( status & (NS_ETH_EINTR_RXNOBUF |
		      NS_ETH_EINTR_RXBUFFUL |
		      NS_ETH_EINTR_RXOVL_DATA |
		      NS_ETH_EINTR_RXOVL_STAT) )
	{
		printk( KERN_CRIT "%s: rx error status: 0x%x\n",
			__FUNCTION__, status );
			
		if( status & NS_ETH_EINTR_RXOVL_DATA )
			ns_eth_recover_rxovl_stall( dev );
	}
			

	mask = NS_ETH_EINTR_RXDONEA;
	for( ring = 0; ring < NS_RX_RINGS; ring++ ) {
		if( status & mask ) {
			ns_eth_rx_process_ring( dev, ring );
			writel( 1 << ring, NS_ETH_RXFREE );
		}

		mask >>= 1;
	}

        spin_unlock( &priv->lock );

	return IRQ_HANDLED;
}

/***********************************************************************
 * @Function: ns_eth_rx_process_ring
 * @Return: nothing
 * @Descr: iterates the ring descriptors and gives frame to kernel. Reallocates
 *         new buffers
 ***********************************************************************/

static inline void ns_eth_rx_process_ring( struct net_device* dev, int ring )
{
	ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
	int i;
	int j;
	volatile int k;
	rx_buffer_desc_t* desc;

	DEBUG_FN( DEBUG_RX | DEBUG_INT | DEBUG_MINOR );

	desc = &priv->rx_desc[ ring*NS_RX_DESC_RING ];

	for( i = ring*NS_RX_DESC_RING; i < (ring+1)*NS_RX_DESC_RING; i++ ) {
		if( desc->s.bits.full ) {
			/*I guess the user would be glad to receive the packet*/

			skb_put( priv->rx_skb[ i ],desc->len-4);/*no checksum*/
			priv->rx_skb[i]->protocol=eth_type_trans(priv->rx_skb[i],
								 dev);
			priv->rx_skb[i]->ip_summed = CHECKSUM_UNNECESSARY;

			consistent_sync( priv->rx_skb[ i ]->data, desc->len,
					 DMA_FROM_DEVICE );

#if 0
			/* This loop is not for cache updates but helps
			 * "breaking" a resource conflict of the ethernet engine
			 * and the CPU. If not present, when receiving many
			 * frames (e.g. nfsmount by initramfs on 2.6.8, we'll
			 * get an error message
			 * ns_eth_rx_int: rx error status: 0x2010000
			 * (RX FIFO overflow). Adjusting AHB Arbiter BRC
			 * Registers (with either CPU or Ethernet more priority)
			 * doesn't help. A udelay(10) at this place would have
			 * the same effect but would run longer I guess.*/
			for( j=0; j < 3; j++ )
				k = ((int*)priv->rx_skb[ i ]->data)[ j ];
#endif

			skb_reserve( priv->rx_skb[ i ], 2 ); /* align 16byte */

			netif_rx( priv->rx_skb[ i ] );
			dev->last_rx = jiffies;

			priv->rx_skb[ i ] = dev_alloc_skb( NS_ETH_FRAME_LEN + 2);

			if( priv->rx_skb[ i ] != NULL )	{
				priv->rx_skb[ i ]->dev = dev;

				/* prepare descriptor */
				priv->rx_desc[ i ].src = (unsigned int*)
					(unsigned int*) virt_to_bus( priv->rx_skb[ i ]->data );
				priv->rx_desc[ i ].len = NS_ETH_FRAME_LEN;
				priv->rx_desc[ i ].s.bits.wrap =
					!((i+1)%(NS_RX_DESC_RING))?1:0;
				priv->rx_desc[ i ].s.bits.intr = 1;
				priv->rx_desc[ i ].s.bits.enable = 1;
				priv->rx_desc[ i ].s.bits.full = 0;
			} else {
				printk( KERN_ERR "%s: out of memory\n",
					__FUNCTION__ );
				priv->rx_desc[ i ].s.bits.enable = 0;
			}
		}
		desc++;
	}
}

static irqreturn_t ns_eth_tx_int( int irq, void* dev_id,struct pt_regs* regs )
{
	struct net_device* dev = (struct net_device*) dev_id;
	ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
	unsigned int status;
	int i;
	tx_buffer_desc_t* tx_buffer_desc;
	char transmitted = 0;
	int first_not_transmitted = -1;
	char kick_dma = 0;

	DEBUG_FN( DEBUG_TX | DEBUG_INT );

        spin_lock( &priv->lock );
	/* acknowledge interrupt */
	status = readl( NS_ETH_EINTR ) & NS_ETH_EINTR_TX_MA;
	writel( status, NS_ETH_EINTR );

	if( status & (NS_ETH_EINTR_TXERR|
		      NS_ETH_EINTR_TXBUFNR))
		printk( KERN_WARNING "%s: error 0x%x\n", __FUNCTION__, status );

	/* check each frame if it has been sent */
	tx_buffer_desc = (tx_buffer_desc_t*) NS_ETH_TXBD + priv->tx_dma_start;
	/* frames sent */
	for( i=priv->tx_dma_start; i<priv->tx_dma_next; i++,tx_buffer_desc++ ) {
		if( priv->tx_skb[ i ] == NULL )
			/* has been processes in a previous round with
			 * some packets untransmitted */
			continue;

		if( !(tx_buffer_desc->s.reg & 0x10000000 ) ) {
			/* we may have had untransmitted the last round */
			if( !(tx_buffer_desc->s.reg & NS_ETH_ETSR_TXOK ) ) {
				printk( KERN_WARNING
					"%s: descr %i empty but not "
					"transmitted ok (0x%x)\n",
					__FUNCTION__, i, tx_buffer_desc->s.reg );
			}

			/* buffer sent */
			dev_kfree_skb_irq( priv->tx_skb[ i ] );
			priv->tx_skb[ i ] = NULL;
			transmitted = 1;
		} else if( first_not_transmitted == -1 ) {
/* @TODO check why it happens and readd message:			DEBUG_ARGS2( DEBUG_ERROR, "retrying %i (0x%x)\n",
   i, tx_buffer_desc->s.reg ); */
			first_not_transmitted = i;
			priv->tx_dma_start = i;	/* don't iterate over known
						 * frames */
			break;
		}
	}

	if( first_not_transmitted != -1 ) {
		/* retry the buggy descriptors */
		kick_dma = 1;
		writel( first_not_transmitted*4, NS_ETH_TXRPTR );
	} else if( transmitted ) {
		if( priv->tx_dma_next == priv->tx_dma_free ) {
			/* we sent all, wind descriptors back to head */
			priv->tx_dma_start = priv->tx_dma_free = 0;
			rmwl( NS_ETH_EGCR1, & ~NS_ETH_EGCR1_ETXDMA );
		} else {
			/* i and tx_buffer_desc point to tx_dma_next+1 */
			priv->tx_dma_start = i;
			writel( i*4, NS_ETH_TXRPTR );
			while( i++ < priv->tx_dma_free ) {
				tx_buffer_desc->s.reg = 0x70000000;
				tx_buffer_desc++;
			}
			(tx_buffer_desc-1)->s.reg = 0xf0000000;

			/* sent the next bunch of frames */
			kick_dma = 1;
		}
		priv->tx_dma_next = priv->tx_dma_free;

		if( netif_queue_stopped( dev ) )
			netif_wake_queue( dev );
	}

	if( kick_dma ) {
		rmwl( NS_ETH_EGCR2, & ~NS_ETH_EGCR2_TCLER );
		rmwl( NS_ETH_EGCR2, | NS_ETH_EGCR2_TCLER );
	}

        spin_unlock( &priv->lock );

	return IRQ_HANDLED;
}

static struct net_device_stats* ns_eth_get_stats( struct net_device* dev )
{
	ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;

	DEBUG_FN( DEBUG_MINOR );

	if( priv != NULL ) {
		/* irq isn't handling statistic registers */
		spin_lock( &priv->lock );

		/* @TODO: not sure whether rx_errors and
		 * tx_errors also contains sub errors */
		ns_eth_stats.rx_bytes += readl( NS_ETH_STAT_RBYT );
		ns_eth_stats.rx_packets += readl( NS_ETH_STAT_RPKT );
		ns_eth_stats.rx_errors += readl( NS_ETH_STAT_RCDE ) +
			readl( NS_ETH_STAT_RCSE ) +
			readl( NS_ETH_STAT_RUND ) +
			readl( NS_ETH_STAT_RFRG ) +
			readl( NS_ETH_STAT_RJBR );
		ns_eth_stats.multicast += readl( NS_ETH_STAT_RMCA );
		ns_eth_stats.rx_length_errors+=readl( NS_ETH_STAT_RFLR );
		ns_eth_stats.rx_over_errors+=readl( NS_ETH_STAT_ROVR );
		ns_eth_stats.rx_crc_errors+=readl( NS_ETH_STAT_RFCS );
		ns_eth_stats.rx_frame_errors+=readl( NS_ETH_STAT_RALN );

		ns_eth_stats.tx_bytes += readl( NS_ETH_STAT_TBYT );
		ns_eth_stats.tx_packets += readl( NS_ETH_STAT_TPKT );
		ns_eth_stats.tx_errors += readl( NS_ETH_STAT_TDFR ) +
			readl( NS_ETH_STAT_TXCL ) +
			readl( NS_ETH_STAT_TJBR ) +
			readl( NS_ETH_STAT_TFCS ) +
			readl( NS_ETH_STAT_TOVR ) +
			readl( NS_ETH_STAT_TUND ) +
			readl( NS_ETH_STAT_TFRG );
		ns_eth_stats.collisions += readl( NS_ETH_STAT_TNCL );
		ns_eth_stats.tx_aborted_errors += readl( NS_ETH_STAT_TXCL );

		spin_unlock( &priv->lock );
	}

	return &ns_eth_stats;
}

static void ns_eth_set_multicast_list( struct net_device* dev )
{
	unsigned int unSAFR;

    	DEBUG_FN( DEBUG_MINOR );


	unSAFR = *get_eth_reg_addr( NS_ETH_SAFR );

	// disable all multicast modes
	unSAFR &= ~( NS_ETH_SAFR_PRO |
		     NS_ETH_SAFR_PRM |
		     NS_ETH_SAFR_PRA );

	if( dev->flags & IFF_PROMISC )
	{
	    DEBUG_ARGS0( DEBUG_MINOR, "Promiscuous mode\n" );
	    *get_eth_reg_addr( NS_ETH_SAFR ) = unSAFR | NS_ETH_SAFR_PRO;
	    return;
	} else if( dev->flags & IFF_ALLMULTI )	{
	    DEBUG_ARGS0( DEBUG_MINOR, "Accept all multicast packets\n" );
	    *get_eth_reg_addr( NS_ETH_SAFR ) = unSAFR |
		NS_ETH_SAFR_BROAD |
		NS_ETH_SAFR_PRM;
	    return;
	} else if( !(dev->flags & IFF_MULTICAST) ) {
	    // TODO: check if broadcasts will be delivered correctly or if they
	    // need multicasts to be enabled
	    DEBUG_ARGS0( DEBUG_MINOR,
			"Accept only direct addressed and broadcast packets\n" );
	    *get_eth_reg_addr( NS_ETH_SAFR ) = unSAFR |
		NS_ETH_SAFR_BROAD;
	    return;
	}

	DEBUG_ARGS0( DEBUG_MINOR,
		    "Accept multicast list\n" );

	// store multicast list of kernel into chip
	ns_load_mca_hash_table( dev->mc_list );

	*get_eth_reg_addr( NS_ETH_SAFR ) = unSAFR |
	    NS_ETH_SAFR_BROAD |
	    NS_ETH_SAFR_PRA;

}

/***********************************************************************
 * MULTICAST ADDRESS HASH TABLE FUNCTIONS
 * Taken from netarm_eth.c and [1]
 ***********************************************************************/


/***********************************************************************
 * @Function: ns_load_mca_hash_table
 * @Return: nothing
 * @Descr: This function generates the multicast address hash table from the
 *         multicast addresses contained in the provided linked list.  This
 *         function loads this table into the Ethernet controller.
 ***********************************************************************/

static void ns_load_mca_hash_table( struct dev_mc_list* pMcList )
{
	unsigned long      auiHashTable[ 2 ];
	struct dev_mc_list* pMcListEntry;
	int                 nBitNumber;

	// Generate the hash table.

	memset( auiHashTable, 0, sizeof( auiHashTable ) );

	for( pMcListEntry = pMcList;
	     pMcListEntry != NULL;
	     pMcListEntry = pMcListEntry->next)
	{
	    DEBUG_ARGS6( DEBUG_MINOR,
			"  multicast address:  %X %X %X %X %X %X\n",
			pMcListEntry->dmi_addr[0], pMcListEntry->dmi_addr[1],
			pMcListEntry->dmi_addr[2], pMcListEntry->dmi_addr[3],
			pMcListEntry->dmi_addr[4], pMcListEntry->dmi_addr[5] );

	    nBitNumber = ns_calculate_hash_bit( pMcListEntry->dmi_addr );
	    ns_set_hash_bit( (unsigned char *) auiHashTable, nBitNumber );
	}
	DEBUG_ARGS2( DEBUG_MINOR,
		    "  hash table:  %lX %lX \n",
		    auiHashTable[ 0 ], auiHashTable[ 1 ] );

	// Load the hash table into the Ethernet controller.

	*get_eth_reg_addr( NS_ETH_HT1 ) = auiHashTable[ 0 ];
	*get_eth_reg_addr( NS_ETH_HT2 ) = auiHashTable[ 1 ];
}

/***********************************************************************
 * @Function: ns_set_hash_bit
 * @Return: nothing
 * @Descr: sets a bit in the hash table
 ***********************************************************************/

static void ns_set_hash_bit( unsigned char* pucHashTable, int nBit )
{
	int nByteIndex;
	int nBitIndex;

	nByteIndex = nBit >> 3;
	nBitIndex = nBit & 7;
	pucHashTable[ nByteIndex ] |= (1 << nBitIndex );
}


/***********************************************************************
 * @Function: ns_calculate_hash_bit
 * @Return: bit position to set in hash table
 * @Descr:  This routine calculates the bit to set in the CRC hash table
 *          to enable receipt of the specified multicast address.
 ***********************************************************************/

static int ns_calculate_hash_bit( unsigned char* pucMac )
{
	unsigned short  unCopyMac[ 3 ];
	unsigned short* punMac;
	unsigned short  unBp;
	unsigned short  unBx;
	unsigned long   ulCrc;
	int             nResult;
	int             nMacWord;
	int             nBitIndex;

	// pucMac may not be 32 bit aligned
	memcpy( unCopyMac, pucMac, sizeof( unCopyMac ) );

	punMac = unCopyMac;
	ulCrc = 0xffffffffL;

	for( nMacWord = 0; nMacWord < 3; nMacWord++ )
	{
	    unBp = *punMac;                            // get word of address
	    punMac++;

	    for( nBitIndex = 0; nBitIndex < 16; nBitIndex++ )
	    {
		unBx = (unsigned short) (ulCrc >> 16);  // get high word of crc
		unBx = ns_rotate(unBx,NS_LEFT,1); // bit 31 to lsb
		unBx ^= unBp;                           // combine with incoming
		ulCrc <<= 1;                           // shift crc left 1 bit
		unBx &= 1;                             // get control bit
		if( unBx )                             // if bit set
		    ulCrc ^= NS_POLYNOMIAL;        // xor crc with polynomial
		ulCrc |= unBx;                         // or in control bit
		unBp = ns_rotate( unBp, NS_RIGHT, 1 );
	    }
	}

	// CRC calculation done.  The 6-bit result resides in bits locations
	// 28:23.

	nResult = (ulCrc >> 23) & 0x3f;

	return nResult;
}


/***********************************************************************
 * @Function: ns_rotate
 * @Return: rotated short
 * @Descr:
 *      rotates an unsigned short a number of bit
 *      reg         value to rotate
 *      direction   NS_LEFT or NS_RIGHT
 *      bits        # of bits to rotate
 ***********************************************************************/

static unsigned short ns_rotate( unsigned short uiReg,
				     int nDirection,
				     int nBits )
{
	int i;

	if( nDirection == NS_LEFT )
	{
	    for( i = 0; i < nBits; i++ )
	    {
		if( uiReg & 0x8000 )            // if high bit set
		    uiReg = (uiReg << 1) | 1;   // shift left and or in 1
		else
		    uiReg <<= 1;
	    }
	} else {                                // if ror
	    for( i = 0; i < nBits; i++ )
	    {
		if( uiReg & 1 )                 // if low bit set
		    uiReg = (uiReg>>1)|0x8000;  // shift right and or in hi bit
		else
		    uiReg >>= 1;
	    }
	}

	return uiReg;
}


static int ns_eth_set_mac_address( struct net_device *dev, void *p )
{
	struct sockaddr *addr = p;

	memcpy(dev->dev_addr, addr, dev->addr_len);

	writel( dev->dev_addr[ 5 ] << 8 |
		dev->dev_addr[ 4 ], NS_ETH_SA1 );
	writel( dev->dev_addr[ 3 ] << 8 |
		dev->dev_addr[ 2 ], NS_ETH_SA2 );
	writel( dev->dev_addr[ 1 ] << 8 |
		dev->dev_addr[ 0 ], NS_ETH_SA3 );

	return 0;
}

static void ns_eth_tx_timeout( struct net_device* dev )
{
	DEBUG_FN( DEBUG_TX | DEBUG_MINOR );
}

static int ns_init_device( struct net_device* dev )
{
	DEBUG_FN( DEBUG_INIT );

	memset( &ns_eth_stats, 0, sizeof( ns_eth_stats ) );

	ether_setup( dev );

	dev->dev_addr[ 5 ] = (readl( NS_ETH_SA1 ) >> 8) & 0xff;
	dev->dev_addr[ 4 ] = (readl( NS_ETH_SA1 )) & 0xff;
	dev->dev_addr[ 3 ] = (readl( NS_ETH_SA2 ) >> 8) & 0xff;
	dev->dev_addr[ 2 ] = (readl( NS_ETH_SA2 )) & 0xff;
	dev->dev_addr[ 1 ] = (readl( NS_ETH_SA3 ) >> 8) & 0xff;
	dev->dev_addr[ 0 ] = (readl( NS_ETH_SA3 )) & 0xff;

	return 0;
}

static void ns_irqs_disable( void )
{
        DEBUG_FN( DEBUG_INIT );

        disable_irq( IRQ_ETH_RX );
        disable_irq( IRQ_ETH_TX );
}

static void ns_irqs_enable( void )
{
        DEBUG_FN( DEBUG_INIT );

        enable_irq( IRQ_ETH_RX );
        enable_irq( IRQ_ETH_TX );
}

/***********************************************************************
 * @Function: ns_eth_reset
 * @Return: 0 on failure otherwise 1
 * @Descr: resets the ethernet interface and the PHY,
 *         performs auto negotiation or fixed modes
 ***********************************************************************/

static int ns_eth_reset( struct net_device* dev )
{
	ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;

	DEBUG_FN( DEBUG_MINOR );

	/* Reset MAC */
	rmwl( NS_ETH_EGCR1, | NS_ETH_EGCR1_MAC_HRST );
	udelay( 5 ); 		/* according to [1], p.322 */
	rmwl( NS_ETH_EGCR1, & ~NS_ETH_EGCR1_MAC_HRST );

	/* reset and initialize PHY */

	rmwl( NS_ETH_MAC1,  & ~NS_ETH_MAC1_SRST );

	/* we don't support hot plugging of PHY, therefore we don't reset
	   phyDetected and nPhyMaxMdioClock here. The risk is if the setting is
	   incorrect the first open
	   may detect the PHY correctly but succeding will fail
	   For reseting the PHY and identifying we have to use the standard
	   MDIO CLOCK value 2.5 MHz only after hardware reset
	   After having identified the PHY we will do faster */

	writel( ns_mii_get_clock_divisor( phy_mdio_max_clk ), NS_ETH_MCFG );

	/* reset PHY */
	ns_mii_write( PHY_COMMON_CTRL, PHY_COMMON_CTRL_RESET );
	ns_mii_write( PHY_COMMON_CTRL, 0 );

	udelay( 350 ); /* [2] p.70 says at least 300us reset recovery time */

	/* MII clock has been setup to default, ns_mii_identify_phy should
	   work for all */

	if( !ns_mii_identify_phy() ) {
	    printk( KERN_ERR DRIVER_NAME ": Unsupported PHY, aborting\n");
	    return 0;
	}

	/* now take the highest MDIO clock possible after detection */
	writel(ns_mii_get_clock_divisor(phy_mdio_max_clk), NS_ETH_MCFG );

	priv->last_link_status = 0xff; /* undefined */

	if((link_mode&FS_EEPROM_AUTONEG_ENABLE_MASK)==FS_EEPROM_AUTONEG_DISABLE)
		/* use parameters defined */
		ns_link_force( dev );
	else
		ns_link_auto_negotiate( dev );

	if( phy_detected == PHY_LXT971A )
		/* set LED2 to link mode */
		ns_mii_write( PHY_LXT971_LED_CFG,
				  PHY_LXT971_LED_CFG_LINK_ACT <<
				  PHY_LXT971_LED_CFG_SHIFT_LED2);

	/* reset interrupt status register */
	writel( NS_ETH_EINTR_RX_MA | NS_ETH_EINTR_TX_MA, NS_ETH_EINTR );

	return 1;
}

/***********************************************************************
 * @Function: ns_eth_enable
 * @Return: nothing
 * @Descr: configures ethernet for link mode, enables FIFO and DMA
 ***********************************************************************/

static void ns_eth_enable( struct net_device* dev )
{
	ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
	int i;

	DEBUG_FN( DEBUG_INIT | DEBUG_MINOR );

	/* enable hardware */

	/* enable receive and transmit FIFO, use 10/100 Mbps MII */
	writel( NS_ETH_EGCR1_ETXWM |
		NS_ETH_EGCR1_ERX |
		NS_ETH_EGCR1_ETX |
		NS_ETH_EGCR1_RXALIGN |
		NS_ETH_EGCR1_PHY_MODE_MII |
		NS_ETH_EGCR1_ITXA,
		NS_ETH_EGCR1 );

	/* EINTR can only be reseted with ERX and ETX on */
	rmwl( NS_ETH_EINTR, &0xffffffff );

	writel( NS_ETH_MAC1_RXEN, NS_ETH_MAC1 );

	/* the linux kernel may give packets < 60 bytes, for example arp */
	writel( NS_ETH_MAC2_CRCEN |
		NS_ETH_MAC2_PADEN,
		NS_ETH_MAC2 );

	/* program the first descriptors for each ring. */
	for( i = 0; i < NS_RX_RINGS; i++ )
		writel((unsigned int) (priv->rx_dma_desc + i*NS_RX_DESC_RING),
			NS_ETH_RXAPTR + i * 4 );

	ns_eth_update_rx_desc( dev );

	rmwl( NS_ETH_EGCR1, | NS_ETH_EGCR1_ERXDMA );

	/* ignore statistics interrupt */
	writel( NS_ETH_EINTR_RXOVL_DATA |
		NS_ETH_EINTR_RXBUFC |
		NS_ETH_EINTR_RXBUFC |
		NS_ETH_EINTR_RXDONEA |
		NS_ETH_EINTR_RXDONEB |
		NS_ETH_EINTR_RXDONEC |
		NS_ETH_EINTR_RXDONED |
		NS_ETH_EINTR_RXNOBUF |
		NS_ETH_EINTR_RXBUFFUL |
		NS_ETH_EINTR_TXBUFNR |
		NS_ETH_EINTR_TXDONE |
		NS_ETH_EINTR_TXERR |
		NS_ETH_EINTR_TXIDLE,
		NS_ETH_EINTREN );

	/* clear statistics */
	writel( NS_ETH_EGCR2_CLRCNT, NS_ETH_EGCR2 );
	/* enable statistics */
	writel( NS_ETH_EGCR2_AUTOZ | NS_ETH_EGCR2_STEN, NS_ETH_EGCR2 );
}

static void ns_eth_update_rx_desc( struct net_device* dev )
{
	int timeout = 20;

	DEBUG_FN( DEBUG_INIT | DEBUG_MINOR );

	rmwl( NS_ETH_EGCR1, | NS_ETH_EGCR1_ERXINIT );
	/* [1] Tab. 221 states less than 5us */
	while( --timeout && !(readl( NS_ETH_EGSR ) & NS_ETH_EGSR_RXINIT))
		/* wait for finish */
		udelay( 1 );

	if( !timeout )
		printk( KERN_WARNING DRIVER_NAME
			":ethernet descriptor update failed\n" );

	/* @TODO do we need to clear RXINIT? */
	rmwl( NS_ETH_EGCR1, & ~NS_ETH_EGCR1_ERXINIT );

	/* all 4 descriptors free */
	writel( 0xf, NS_ETH_RXFREE );
}

/***********************************************************************
 * @Function: ns_link_force
 * @Return: void
 * @Descr: configures eth and MII to use the link mode defined in
 *         ucLinkMode
 ***********************************************************************/

static void ns_link_force( struct net_device* dev )
{
	ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
	unsigned short control;

	DEBUG_FN( DEBUG_LINK );

	control = ns_mii_read( PHY_COMMON_CTRL );
	control &= ~( PHY_COMMON_CTRL_SPD_MA |
			PHY_COMMON_CTRL_AUTO_NEG |
			PHY_COMMON_CTRL_DUPLEX );

	priv->last_link_status = 0;

	if( ( link_mode & FS_EEPROM_AUTONEG_SPEED_MASK ) ==
	    FS_EEPROM_AUTONEG_SPEED_100 ) {
	    control |= PHY_COMMON_CTRL_SPD_100;
	    priv->last_link_status |= PHY_LXT971_STAT2_100BTX;
	} else
	    control |= PHY_COMMON_CTRL_SPD_10;

	if( ( link_mode & FS_EEPROM_AUTONEG_DUPLEX_MASK ) ==
	    FS_EEPROM_AUTONEG_DUPLEX_FULL ) {
	    control |= PHY_COMMON_CTRL_DUPLEX;
	    priv->last_link_status |= PHY_LXT971_STAT2_DUPLEX_MODE;
	}

	ns_mii_write( PHY_COMMON_CTRL, control );

	ns_link_print_changed( dev );
	ns_link_update_egcr( dev );
}

/***********************************************************************
 * @Function: ns_link_auto_negotiate
 * @Return: void
 * @Descr: performs auto-negotation of link.
 ***********************************************************************/

static void ns_link_auto_negotiate( struct net_device* dev )
{
	unsigned long start_jiffies;
	unsigned short status;

	DEBUG_FN( DEBUG_LINK );

	status = ns_mii_read( PHY_COMMON_STAT );
	if( ( status &
	      (PHY_COMMON_STAT_AN_COMP | PHY_COMMON_STAT_LNK_STAT)) ==
	    (PHY_COMMON_STAT_AN_COMP | PHY_COMMON_STAT_LNK_STAT) ) {

		ns_link_print_changed( dev );
		ns_link_update_egcr( dev );
		return;
	}

	/* run auto-negotation */
	/* define what we are capable of */
	ns_mii_write( PHY_COMMON_AUTO_ADV,
		      PHY_COMMON_AUTO_ADV_100BTXFD |
		      PHY_COMMON_AUTO_ADV_100BTX |
		      PHY_COMMON_AUTO_ADV_10BTFD |
		      PHY_COMMON_AUTO_ADV_10BT |
		      PHY_COMMON_AUTO_ADV_802_3 );
	/* start auto-negotiation */
	ns_mii_write( PHY_COMMON_CTRL,
		      PHY_COMMON_CTRL_AUTO_NEG |
		      PHY_COMMON_CTRL_RES_AUTO );

	/* wait for completion */

	start_jiffies = jiffies;
	while( jiffies < start_jiffies + NS_MII_NEG_DELAY ) {
		status = ns_mii_read( PHY_COMMON_STAT );
		if( ( status &
		      (PHY_COMMON_STAT_AN_COMP | PHY_COMMON_STAT_LNK_STAT)) ==
		    (PHY_COMMON_STAT_AN_COMP | PHY_COMMON_STAT_LNK_STAT) ) {
			/* lucky we are, auto-negotiation succeeded */
			ns_link_print_changed( dev );
			ns_link_update_egcr( dev );
			return;
		}
	}

	printk( KERN_WARNING DRIVER_NAME
		":auto-negotiation timed out, forcing 10Mbps/Half\n" );

	/* force mode */
	link_mode = FS_EEPROM_AUTONEG_SPEED_10 | FS_EEPROM_AUTONEG_DUPLEX_HALF;
	ns_link_force( dev );
}

/***********************************************************************
 * @Function: ns_link_update_egcr
 * @Return: void
 * @Descr: updates the EGCR and MAC2 link status after mode change or
 *         auto-negotation
 ***********************************************************************/

static void ns_link_update_egcr( struct net_device* dev )
{
	ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
	unsigned int egcr;
	unsigned int mac2;
	unsigned int ipgt;

	DEBUG_FN( DEBUG_LINK );

	egcr = readl( NS_ETH_EGCR1 );
	mac2 = readl( NS_ETH_MAC2 );
	ipgt = readl( NS_ETH_IPGT ) & ~NS_ETH_IPGT_MA;

	mac2 &= ~NS_ETH_MAC2_FULLD;
	if( (priv->last_link_status & PHY_LXT971_STAT2_DUPLEX_MODE)
	    == PHY_LXT971_STAT2_DUPLEX_MODE ) {
		mac2 |= NS_ETH_MAC2_FULLD;
		ipgt |= 0x15;	/* see [1] p. 339 */
	} else
		ipgt |= 0x12;	/* see [1] p. 339 */

	writel( mac2, NS_ETH_MAC2 );
	writel( egcr, NS_ETH_EGCR1 );
	writel( ipgt, NS_ETH_IPGT );
}

/***********************************************************************
 * @Function: ns_link_print_changed
 * @Return: void
 * @Descr: checks whether the link status has changed and if so prints
 *         the new mode
 ***********************************************************************/

static void ns_link_print_changed( struct net_device* dev )
{
	ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
	unsigned short status;
	unsigned short control;

	DEBUG_FN( DEBUG_LINK );

	control = ns_mii_read( PHY_COMMON_CTRL );

	if( (control & PHY_COMMON_CTRL_AUTO_NEG) == PHY_COMMON_CTRL_AUTO_NEG) {
		/* PHY_COMMON_STAT_LNK_STAT is only set on autonegotiation */
		status = ns_mii_read( PHY_COMMON_STAT );

		if( !( status & PHY_COMMON_STAT_LNK_STAT) ) {
			printk( KERN_WARNING DRIVER_NAME ": link down\n");
			netif_carrier_off( dev );
		} else {
			if( !netif_carrier_ok( dev ) )
				/* we are online again, enable it */
				netif_carrier_on( dev );

			if( phy_detected == PHY_LXT971A ) {
				status = ns_mii_read( PHY_LXT971_STAT2 );
				status &= (PHY_LXT971_STAT2_100BTX |
					     PHY_LXT971_STAT2_DUPLEX_MODE |
					     PHY_LXT971_STAT2_AUTO_NEG );

				/* mask out all uninteresting parts */
			}
			/* other PHYs must store there link information in
			   uiStatus as PHY_LXT971 */
		}
	} else {
		/* mode has been forced, so uiStatus should be the same as the
		   last link status, enforce printing */
		status = priv->last_link_status;
		priv->last_link_status = 0xff;
	}

	if( status != priv->last_link_status ) {
		/* save current link status */
		priv->last_link_status = status;

		/* print new link status */

		printk( KERN_INFO DRIVER_NAME
			": link mode %i Mbps %s duplex %s\n",
			(status & PHY_LXT971_STAT2_100BTX) ? 100 : 10,
			(status & PHY_LXT971_STAT2_DUPLEX_MODE)?"full":"half",
			(status & PHY_LXT971_STAT2_AUTO_NEG)?"(auto)":"" );
	}
}

/***********************************************************************
 * the MII low level stuff
 ***********************************************************************/

/***********************************************************************
 * @Function: ns_mii_identify_phy
 * @Return: 1 if supported PHY has been detected otherwise 0
 * @Descr: checks for supported PHY and prints the IDs.
 ***********************************************************************/

static char ns_mii_identify_phy( void )
{
	unsigned short id1;
	unsigned short id2;
	unsigned char* name;

	char ret = 0;

	DEBUG_FN( DEBUG_MII );

	for( NS_ETH_PHY_ADDRESS = 0; NS_ETH_PHY_ADDRESS < 31;
	     NS_ETH_PHY_ADDRESS++ ) {
		id1 = ns_mii_read(PHY_COMMON_ID1);
		phy_detected = (phy_type_t) id1;

		switch( phy_detected ) {
		    case PHY_LXT971A:
			name = "LXT971A";
			id2 = ns_mii_read( PHY_COMMON_ID2 );
			phy_mdio_max_clk = PHY_LXT971_MDIO_MAX_CLK;
			ret = 1;
			break;
		    case PHY_ICS1893BK:
		    	name = "ICS1893BK";
			id2 = ns_mii_read( PHY_COMMON_ID2 );
			phy_mdio_max_clk = PHY_ICS1893_MDIO_MAX_CLK;
			ret = 1;
			break;
		    case PHY_NONE:
		    default:
			/* in case id1 == 0 && id2 == 0 we may have the wrong
			   address or reset sets the wrong NS_ETH_MCFG_CLKS */

			id2 = 0;
			name = "unknown";
			phy_mdio_max_clk = PHY_MDIO_MAX_CLK;
			phy_detected = PHY_NONE;
		}
		if( ret )
			break;
	}

	printk( KERN_INFO DRIVER_NAME
		": PHY (0x%x, 0x%x) = %s detected at 0x%x\n",
		id1, id2, name, NS_ETH_PHY_ADDRESS );

	return ret;
}


/***********************************************************************
 * @Function: ns_mii_read
 * @Return: the data read from PHY register uiRegister
 * @Descr: the data read may be invalid if timed out. If so, a message
 *         is printed but the invalid data is returned.
 *         The fixed device address is being used.
 ***********************************************************************/

static unsigned short ns_mii_read( unsigned short reg )
{
	DEBUG_FN( DEBUG_MII_LOW );

	/* write MII register to be read */
	writel( NS_ETH_PHY_ADDRESS<<8|reg, NS_ETH_MADR );
	writel( NS_ETH_MCMD_READ, NS_ETH_MCMD );

	if( !ns_mii_poll_busy() )
	    printk( KERN_WARNING DRIVER_NAME
		    ": MII still busy in read\n" );
	/* continue to read */

	writel( 0, NS_ETH_MCMD );

	return (unsigned short) (readl( NS_ETH_MRDD ) );
}


/***********************************************************************
 * @Function: ns_mii_write
 * @Return: nothing
 * @Descr: writes the data to the PHY register. In case of a timeout,
 *         no special handling is performed but a message printed
 *         The fixed device address is being used.
 ***********************************************************************/

static void ns_mii_write( unsigned short reg, unsigned short data )
{
	DEBUG_FN( DEBUG_MII_LOW );

	/* write MII register to be written */
	writel( NS_ETH_PHY_ADDRESS << 8 | reg, NS_ETH_MADR );
	writel( data, NS_ETH_MWTD );

	if( !ns_mii_poll_busy() )
		printk( KERN_WARNING DRIVER_NAME
			": MII still busy in write\n");
}


/***********************************************************************
 * @Function: ns_mii_get_clock_divisor
 * @Return: the clock divisor that should be used in NS_ETH_MCFG_CLKS
 * @Descr: if no clock divisor can be calculated for the
 *         current SYSCLK and the maximum MDIO Clock, a warning is printed
 *         and the greatest divisor is taken
 ***********************************************************************/

#ifndef AHB_CLK_FREQ
#define AHB_CLK_FREQ ns9xxx_ahbclock()
#endif

static unsigned int ns_mii_get_clock_divisor( unsigned int max_mdio_clk )
{
	struct
	{
		unsigned int clk_div;
		unsigned int clks; /* field for NS_ETH_MCFG_CLKS */
	} PHYClockDivisors[] = {
		{  4, NS_ETH_MCFG_CLKS_4 },
		{  6, NS_ETH_MCFG_CLKS_6 },
		{  8, NS_ETH_MCFG_CLKS_8 },
		{ 10, NS_ETH_MCFG_CLKS_10 },
		{ 20, NS_ETH_MCFG_CLKS_20 },
		{ 30, NS_ETH_MCFG_CLKS_30 },
		{ 40, NS_ETH_MCFG_CLKS_40 }
	};

	int i;
	int size = sizeof( PHYClockDivisors ) / sizeof( PHYClockDivisors[ 0 ] );
	unsigned int clks = NS_ETH_MCFG_CLKS_40; /* defaults to greatest div */

	DEBUG_FN( DEBUG_INIT );

	for( i = 0; i < size; i++ ) {
		/* find first sysclock divisor that isn't higher than 2.5 MHz
		   clock */
		if( AHB_CLK_FREQ/PHYClockDivisors[i].clk_div <= max_mdio_clk ) {
			clks = PHYClockDivisors[ i ].clks;
			break;
		}
	}

	DEBUG_ARGS2( DEBUG_INIT,
		     "Taking MDIO Clock bit mask 0x%0x for max clock %i\n",
		     clks,
		     max_mdio_clk );

	/* return greatest divisor */
	return clks;
}

/***********************************************************************
 * @Function: ns_mii_poll_busy
 * @Return: 0 if timed out otherwise the remaing timeout
 * @Descr: waits until the MII has completed a command or it times out
 *         code may be interrupted by hard interrupts.
 *         It is not checked what happens on multiple actions when
 *         the first is still being busy and we timeout.
 ***********************************************************************/

static unsigned int ns_mii_poll_busy( void )
{
	unsigned int timeout = 10000;

	DEBUG_FN( DEBUG_MII_LOW );

	while( (( readl( NS_ETH_MIND ) & NS_ETH_MIND_BUSY)
		== NS_ETH_MIND_BUSY ) && timeout )
		timeout--;

	return timeout;
}

static int __init ns_eth_init_module( void )
{
	int result = 0;

#ifdef CONFIG_ARCH_NS9360
#  include <asm/arch/ns9750_bbus.h>
	int i;
	printk("Setting GPIOs for NS9360 ethernet\n");
#  ifdef CONFIG_MACH_CC9C
	/*GPIO65 is used as BOOTMUX and must be low to
	  enable user buffers
	  -->PHY_INT cannot be used
	  */
	for( i=50; i<=64; i++)
		set_gpio_cfg( i, NS_GPIO_CFG_FUNC_0 );
#  else
#    ifdef CONFIG_MACH_CCW9C
	/*GPIO58, 65 are used by the wireless device */
	for( i=50; i<58; i++)
		set_gpio_cfg( i, NS_GPIO_CFG_FUNC_0 );
	for( i=59; i<65; i++)
		set_gpio_cfg( i, NS_GPIO_CFG_FUNC_0 );
    #else 
	for( i=50; i<=65; i++)
		set_gpio_cfg( i, NS_GPIO_CFG_FUNC_0 );

    #endif /*CONFIG_MACH_CCW9C*/
  #endif /*CONFIG_MACH_CC9C*/
#endif /*CONFIG_ARCH_NS9360*/

#ifdef CONFIG_ARCH_NS9XXX
	if (processor_is_ns9360()) {
		/* XXX: create and use gpio api */
		/* configure gpios 50 to 64 as func0 */
//		*(volatile u32 *)0x90600028 &= 0x000000ff;
//		*(volatile u32 *)0x90600100 = 0x00000000;
//		*(volatile u32 *)0x90600104 &= 0xfffffff0;
	}
#endif

	if( ( result = register_netdev( &ns_eth_device ) ) != 0 )
		return result;

	printk( version );

	return 0;
}

static void __exit ns_eth_cleanup_module( void )
{
	unregister_netdev( &ns_eth_device );
}

module_init( ns_eth_init_module );
module_exit( ns_eth_cleanup_module );

MODULE_AUTHOR( "Markus Pietrek <mpietrek@fsforth.de>" );
MODULE_DESCRIPTION( "Ethernet driver for NS9xx0" );
MODULE_LICENSE( "GPL" );
