/*
 * linux/drivers/pcmcia/pxa2xx_ccxp.c
 *
 * Digi ConnectCore XP PCMCIA specific routines.
 *
 * @Author: Bernd Westermann
 * @References: [1] lubbock.h
 *
 * Copyright (C) 2004 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 */

/***********************************************************************
 *  @History:
 *     2005/07/08 : initial Version
 *     2005/08/14 : renaming to ccxp270
 ***********************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/device.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/platform_device.h>

#include <asm/hardware.h>
#include <asm/mach-types.h>
#include <asm/arch/ccxp.h>
#include <asm/irq.h>

#include "soc_common.h"

static struct pcmcia_irqs irqs[3 * 2] = { {0, }, };
static int n_irqs = 0;
static int rdy_gpios[2] = { CCXP_RDY1_GPIO, CCXP_RDY2_GPIO };
static int cd_gpios[2] = { CCXP_CD1_GPIO, CCXP_CD2_GPIO };
static int rst_gpios[2] = { CCXP_RST1_GPIO, CCXP_RST2_GPIO };
static int bvd1_gpios[2] = { -1, -1 };
static int bvd2_gpios[2] = { -1, -1 };

static void ccxp_pcmcia_reset_socket(int skt, int reset)
{
	static int reset_flags = 0;

	/*printk("%s: socket=%d, reset_flags=%08x\n", __FUNCTION__, skt, reset_flags);*/
	if (reset) {
		if ((reset_flags & (1 << skt)) == 0) {
			/*printk("%s: Resetting PCMCIA slot %d via GPIO%d\n",
				__FUNCTION__, skt, rst_gpios[skt]);*/
			gpio_set_bit(rst_gpios[skt]);
			reset_flags |= (1 << skt);
		}
	} else {
		if ((reset_flags & (1 << skt)) != 0) {
			/*printk("%s: Releasing reset for PCMCIA slot %d via GPIO%d\n",
				__FUNCTION__, skt, rst_gpios[skt]);*/
			gpio_clr_bit(rst_gpios[skt]);
			reset_flags &= ~(1 << skt);
		}
	}
}

static int pxa_ccxp_pcmcia_hw_init(struct soc_pcmcia_socket *skt)
{
	int ret = 0;
	int i;

	/*
	 * Setup default state of GPIO outputs
	 * before we enable them as outputs.
	 */
	GPSR(GPIO48_nPOE) =
		GPIO_bit(GPIO48_nPOE) |
		GPIO_bit(GPIO49_nPWE) |
		GPIO_bit(GPIO50_nPIOR) |
		GPIO_bit(GPIO51_nPIOW) |
		GPIO_bit(GPIO85_nPCE_1) |
		GPIO_bit(GPIO54_nPCE_2);

	pxa_gpio_mode(GPIO48_nPOE_MD);
	pxa_gpio_mode(GPIO49_nPWE_MD);
	pxa_gpio_mode(GPIO50_nPIOR_MD);
	pxa_gpio_mode(GPIO51_nPIOW_MD);
	pxa_gpio_mode(GPIO85_nPCE_1_MD);
	pxa_gpio_mode(GPIO54_nPCE_2_MD);
	pxa_gpio_mode(GPIO104_pSKTSEL_MD); /*ist 104 war 79*/
	pxa_gpio_mode(GPIO55_nPREG_MD);
	pxa_gpio_mode(GPIO56_nPWAIT_MD);
	pxa_gpio_mode(GPIO57_nIOIS16_MD);


	/*printk("%s: socket %d\n", __FUNCTION__, skt->nr);*/
	if (skt->irq == NO_IRQ) {
		if (skt->nr < NUMBER_OF_CCXP_PCMCIA_SLOTS && VALID_SLOT(skt->nr)) {
			skt->irq = IRQ_GPIO(rdy_gpios[skt->nr]);
			/*printk("%s: RDY IRQ for socket %d is: %d\n", __FUNCTION__,
				skt->nr, skt->irq);*/

		} else {
			printk("%s: IRQ info for socket %d not available\n", __FUNCTION__,
				skt->nr);
			ret = -1;
		}
	}

	/*printk("%s: Requesting IRQs for socket: %d\n", __FUNCTION__, skt->nr);*/
	for (i = 0; i < n_irqs; i++) {
		if (irqs[i].sock == skt->nr) {
			/*printk("%s: %s IRQ=%d\n", __FUNCTION__, irqs[i].str, irqs[i].irq);*/
		}
	}
	ret = soc_pcmcia_request_irqs(skt, irqs, n_irqs);
	/*printk("%s: soc_pcmcia_request_irqs(%d) returned %d\n", __FUNCTION__, skt->nr, ret);*/

	return ret;
}

static void pxa_ccxp_pcmcia_hw_shutdown(struct soc_pcmcia_socket *skt)
{
	if (VALID_SLOT(skt->nr)) {
		/*printk("%s: Shutting down socket %d\n", __FUNCTION__, skt->nr);*/
		soc_pcmcia_free_irqs(skt, irqs, n_irqs);
	} else {
		printk("%s: Socket %d not under PCMCIA control\n", __FUNCTION__, skt->nr);
	}
}

static int pxa_ccxp_pcmcia_card_detect(int slot)
{
	int cd = 0;

	cd = gpio_tst_bit(cd_gpios[slot]) == 0;

	return cd;
}

static void pxa_ccxp_pcmcia_socket_state(struct soc_pcmcia_socket *skt, struct pcmcia_state *state)
{
	state->detect = pxa_ccxp_pcmcia_card_detect(skt->nr);

	if (VALID_SLOT(skt->nr)) {
		state->ready  = VALID_SLOT(skt->nr) ? gpio_tst_bit(rdy_gpios[skt->nr]) : 0;
		if (bvd1_gpios[skt->nr] >= 0) {
			state->bvd1 = gpio_tst_bit(bvd1_gpios[skt->nr]) ? 1 : 0;
		} else {
			state->bvd1 = 0;
		}
		if (bvd2_gpios[skt->nr] >= 0) {
			state->bvd2 = gpio_tst_bit(bvd2_gpios[skt->nr]) ? 1 : 0;
		} else {
			state->bvd2 = 0;
		}
		state->wrprot = 0;
		state->vs_3v  = 1;
		state->vs_Xv  = 0;
	} else {
		printk("%s: Socket %d not under PCMCIA control\n", __FUNCTION__, skt->nr);
		state->ready  = 0;
	}
}

static int pxa_ccxp_pcmcia_configure_socket(struct soc_pcmcia_socket *skt,
					    const socket_state_t *state)
{
	int ret = 0;

	/*printk("%s: sock: %d reset: %d vcc: %d vpp: %d out: %d spkr: %d\n", __FUNCTION__,
		skt->nr, state->flags & SS_RESET, state->Vcc, state->Vpp,
		state->flags & SS_OUTPUT_ENA, state->flags & SS_SPKR_ENA);*/

	if (skt->nr >= NUMBER_OF_CCXP_PCMCIA_SLOTS) {
		printk("%s: socket number %d out of range 0..%d\n", __FUNCTION__,
			skt->nr, NUMBER_OF_CCXP_PCMCIA_SLOTS-1);
		return -1;
	}

	if (!VALID_SLOT(skt->nr)) {
		printk("%s: Socket %d not under PCMCIA control\n", __FUNCTION__, skt->nr);
		return 0;
	}

	// no suport for power switching on CCXP/KARO boards
	switch (state->Vcc) {
	case 0:
		/*printk("%s: Vcc off not supported\n", __FUNCTION__);*/
		break;
	case 33:
		/*printk("%s: 3.3V always on\n", __FUNCTION__);*/
		break;
	case 50:
		/*printk("%s: 5V slots not supported\n", __FUNCTION__);*/
		ret = -1;
		break;
	default:
		printk(KERN_ERR "%s: unrecognized Vcc %u\n", __FUNCTION__, state->Vcc);
	}

	switch (state->Vpp) {
	case 0:
		/*printk("%s: Vpp off not supported\n", __FUNCTION__);*/
		break;
	case 33:
		break;
	case 120:
		/*printk("%s: Vpp not supported\n", __FUNCTION__);*/
		break;
	default:
		printk(KERN_ERR "%s: unrecognized Vpp %u\n", __FUNCTION__, state->Vpp);
	}

	ccxp_pcmcia_reset_socket(skt->nr, state->flags & SS_RESET);

	return 0;
}

static void pxa_ccxp_pcmcia_socket_init(struct soc_pcmcia_socket *skt)
{
	/*printk("%s: Enable IRQs for socket %d\n", __FUNCTION__, skt->nr);*/
	if (VALID_SLOT(skt->nr)) {
		soc_pcmcia_enable_irqs(skt, irqs, n_irqs);
	} else {
		printk("%s: Socket %d not under PCMCIA control\n", __FUNCTION__, skt->nr);
	}
}

static void pxa_ccxp_pcmcia_socket_suspend(struct soc_pcmcia_socket *skt)
{
	/*printk("%s: Disable IRQs for socket %d\n", __FUNCTION__, skt->nr);*/
	if (VALID_SLOT(skt->nr)) {
		soc_pcmcia_disable_irqs(skt, irqs, n_irqs);
	} else {
		printk("%s: Socket %d not under PCMCIA control\n", __FUNCTION__, skt->nr);
	}
}

static struct pcmcia_low_level pxa_ccxp_pcmcia_ops = {
	.owner			= THIS_MODULE,
	.hw_init		= pxa_ccxp_pcmcia_hw_init,
	.hw_shutdown		= pxa_ccxp_pcmcia_hw_shutdown,
	.socket_state		= pxa_ccxp_pcmcia_socket_state,
	.configure_socket	= pxa_ccxp_pcmcia_configure_socket,
	.socket_init		= pxa_ccxp_pcmcia_socket_init,
	.socket_suspend		= pxa_ccxp_pcmcia_socket_suspend,
};

static void pxa_ccxp_release(struct device *dev)
{

}

struct platform_device ccxp_pcmcia_device = {
	.name = "pxa2xx-pcmcia",
	.dev = {
		.platform_data = &pxa_ccxp_pcmcia_ops,
		.release = pxa_ccxp_release,
	},
};

static int __init pcmcia_ccxp_init(void)
{
	int ret = -ENODEV;
	int i;

	if (!machine_is_ccxp270()) {
		return -ENODEV;
	}

	/*printk("%s: Initializing platform device\n", __FUNCTION__);*/

	for (i = 0; i < 2; i++) {
		if (VALID_SLOT(i)) {
			irqs[n_irqs].sock = i;
			irqs[n_irqs].str = "pcmcia CD";
			irqs[n_irqs].irq = IRQ_GPIO(cd_gpios[i]);
			n_irqs++;
			/*printk("%s: rdy[%d] = GPIO%d, IRQ(%d)\n", __FUNCTION__,
				i, rdy_gpios[i], IRQ_GPIO(rdy_gpios[i]));
			printk("%s: cd[%d]  = GPIO%d, IRQ(%d)\n", __FUNCTION__,
				i, cd_gpios[i], IRQ_GPIO(cd_gpios[i]));
			printk("%s: Slot %d is under PCMCIA control\n", __FUNCTION__, i);*/
		} else {
			printk("%s: Slot %d not under PCMCIA control\n", __FUNCTION__, i);
		}
	}

	pxa_ccxp_pcmcia_ops.nr = n_irqs;

	/*printk("%s: Registering platform device '%s'\n", __FUNCTION__, ccxp_pcmcia_device.name);*/
	ret = platform_device_register(&ccxp_pcmcia_device);
	/*printk("%s: platform_device_register returned %d\n", __FUNCTION__, ret);*/

	return ret;
}

static void __exit pcmcia_ccxp_exit(void)
{
	platform_device_unregister(&ccxp_pcmcia_device);
}

module_init(pcmcia_ccxp_init);
module_exit(pcmcia_ccxp_exit);
MODULE_LICENSE("GPL");
