/* ------------------------------------------------------------------------- */
/* 									     */
/* spi.h - definitions for the spi-bus interface			     */
/* 									     */
/* ------------------------------------------------------------------------- */
/*   Copyright (C) 2005 Guillaume GOURAT
	 Based on linux i2c bus driver.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		     */
/* ------------------------------------------------------------------------- */

#ifndef _LINUX_SPI_H
#define _LINUX_SPI_H

#include <linux/module.h>
#include <linux/types.h>
#include "spi-id.h"
#include <linux/device.h>	/* for struct device */
#include <asm/semaphore.h>

/* --- General options ------------------------------------------------	*/

struct spi_msg;
struct spi_algorithm;
struct spi_adapter;
struct spi_client;
struct spi_driver;
struct spi_client_address_data;

/*
 * The master routines are the ones normally used to transmit data to devices
 * on a bus (or/and read from them). Apart from these basic transfer functions
 * to transmit one message at a time, a more complex version can be used to
 * transmit an arbitrary number of messages without interruption.
 */
extern int spi_master_send(struct spi_client *,const char* ,int);
extern int spi_master_recv(struct spi_client *,char* ,int);
extern int spi_master_xfer(struct spi_client *,const char*, char* ,int);

/* Transfer num messages.
 */
extern int spi_transfer(struct spi_adapter *adap, struct spi_msg msg[],int num);

/*
 * A driver is capable of handling one or more physical devices present on
 * SPI adapters. This information is used to inform the driver of adapter
 * events.
 */

struct spi_driver {
	struct module *owner;
	char name[32];
	int id;
	unsigned int class;
	unsigned int flags;		/* div., see below		*/

	/* Notifies the driver that a new bus has appeared. This routine
	 * can be used by the driver to test if the bus meets its conditions
	 * & seek for the presence of the chip(s) it supports. If found, it
	 * registers the client(s) that are on the bus to the spi admin. via
	 * spi_attach_client.
	 */
	int (*attach_adapter)(struct spi_adapter *);
	int (*detach_adapter)(struct spi_adapter *);

	/* tells the driver that a client is about to be deleted & gives it
	 * the chance to remove its private data. Also, if the client struct
	 * has been dynamically allocated by the driver in the function above,
	 * it must be freed here.
	 */
	int (*detach_client)(struct spi_client *);

	/* a ioctl like command that can be used to perform specific functions
	 * with the device.
	 */
	int (*command)(struct spi_client *client,unsigned int cmd, void *arg);

	struct device_driver driver;
	struct list_head list;
};
#define to_spi_driver(d) container_of(d, struct spi_driver, driver)

extern struct bus_type spi_bus_type;

#define SPI_NAME_SIZE	40

/*
 * spi_client identifies a single device (i.e. chip) that is connected to an
 * spi bus. The behaviour is defined by the routines of the driver. This
 * function is mainly used for lookup & other admin. functions.
 */
struct spi_client {
	int id;
	unsigned int flags;		/* div., see below		*/

	void (*select)(struct spi_client *client, int select);
	void (*release)(struct spi_client *client);

	struct spi_adapter *adapter;	/* the adapter we sit on	*/
	struct spi_driver *driver;	/* and our access routines	*/
	int usage_count;		/* How many accesses currently  */
					/* to the client		*/
	struct device dev;		/* the device structure		*/
	struct list_head list;
	char name[SPI_NAME_SIZE];
	struct completion released;
	int spi_chan;			/* SPI channel to be used 	*/
};
#define to_spi_client(d) container_of(d, struct spi_client, dev)

static inline void *spi_get_clientdata (struct spi_client *dev)
{
	return dev_get_drvdata (&dev->dev);
}

static inline void spi_set_clientdata (struct spi_client *dev, void *data)
{
	dev_set_drvdata (&dev->dev, data);
}

#define SPI_DEVNAME(str)	.name = str

static inline char *spi_clientname(struct spi_client *c)
{
	return &c->name[0];
}

/*
 * The following structs are for those who like to implement new bus drivers:
 * spi_algorithm is the interface to a class of hardware solutions which can
 * be addressed using the same bus algorithms - i.e. bit-banging or the PCF8584
 * to name two of the most common.
 */
struct spi_algorithm {
	char name[32];				/* textual description 	*/
	unsigned int id;

	/* If an adapter algorithm can't to SPI-level access, set master_xfer
	   to NULL. If an adapter algorithm can do SMBus access, set
	   smbus_xfer. If set to NULL, the SMBus protocol is simulated
	   using common SPI messages */
	int (*master_xfer)(struct spi_adapter *adap,struct spi_msg msgs[],
	                   int num);

	/* --- ioctl like call to set div. parameters. */
	int (*algo_control)(struct spi_adapter *, unsigned int, unsigned long);

	/* To determine what the adapter supports */
	u32 (*functionality) (struct spi_adapter *);
};

/*
 * spi_adapter is the structure used to identify a physical spi bus along
 * with the access algorithms necessary to access it.
 */
struct spi_adapter {
	struct module *owner;
	unsigned int id;/* == is algo->id | hwdep.struct->id, 		*/
			/* for registered values see below		*/
	unsigned int class;
	struct spi_algorithm *algo;/* the algorithm to access the bus	*/
	void *algo_data;

	/* --- administration stuff. */
	int (*client_register)(struct spi_client *);
	int (*client_unregister)(struct spi_client *);

	/* data fields that are valid for all devices	*/
	struct semaphore bus_lock;
	struct semaphore clist_lock;

	int timeout;
	int retries;
	struct device dev;		/* the adapter device */
	struct class_device class_dev;	/* the class device */

#ifdef CONFIG_PROC_FS
	/* No need to set this when you initialize the adapter          */
	int inode;
#endif /* def CONFIG_PROC_FS */

	int nr;
	struct list_head clients;
	struct list_head list;
	char name[SPI_NAME_SIZE];
	struct completion dev_released;
	struct completion class_dev_released;
};
#define dev_to_spi_adapter(d) container_of(d, struct spi_adapter, dev)
#define class_dev_to_spi_adapter(d) container_of(d, struct spi_adapter, class_dev)

static inline void *spi_get_adapdata (struct spi_adapter *dev)
{
	return dev_get_drvdata (&dev->dev);
}

static inline void spi_set_adapdata (struct spi_adapter *dev, void *data)
{
	dev_set_drvdata (&dev->dev, data);
}

/*flags for the driver struct: */
#define SPI_DF_NOTIFY	0x01		/* notify on bus (de/a)ttaches 	*/
#if 0
/* this flag is gone -- there is a (optional) driver->detach_adapter
 * callback now which can be used instead */
# define SPI_DF_DUMMY	0x02
#endif

/*flags for the client struct: */
#define SPI_CLIENT_ALLOW_USE		0x01	/* Client allows access */
#define SPI_CLIENT_ALLOW_MULTIPLE_USE 	0x02  	/* Allow multiple access-locks */
						/* on an spi_client */

/* spi adapter classes (bitmask) */
#define SPI_CLASS_HWMON		(1<<0)	/* lm_sensors, ... */
#define SPI_CLASS_TV_ANALOG	(1<<1)	/* bttv + friends */
#define SPI_CLASS_TV_DIGITAL	(1<<2)	/* dvb cards */
#define SPI_CLASS_DDC		(1<<3)	/* spi-matroxfb ? */
#define SPI_CLASS_CAM_ANALOG	(1<<4)	/* camera with analog CCD */
#define SPI_CLASS_CAM_DIGITAL	(1<<5)	/* most webcams */
#define SPI_CLASS_SOUND		(1<<6)	/* sound devices */
#define SPI_CLASS_ALL		(UINT_MAX) /* all of the above */

/* spi_client_address_data is the struct for holding default client
 * addresses for a driver and for the parameters supplied on the
 * command line
 */
struct spi_client_address_data {
	unsigned short *normal_spi;
	unsigned short *normal_spi_range;
	unsigned short *probe;
	unsigned short *probe_range;
	unsigned short *ignore;
	unsigned short *ignore_range;
	unsigned short *force;
};

/* Internal numbers to terminate lists */
#define SPI_CLIENT_END		0xfffe
#define SPI_CLIENT_ISA_END	0xfffefffe

/* The numbers to use to set SPI bus address */
#define ANY_SPI_BUS		0xffff
#define ANY_SPI_ISA_BUS		9191

/* The length of the option lists */
#define SPI_CLIENT_MAX_OPTS 48

/* ----- functions exported by spi.o */

/* administration...
 */
extern int spi_add_adapter(struct spi_adapter *);
extern int spi_del_adapter(struct spi_adapter *);

extern int spi_add_driver(struct spi_driver *);
extern int spi_del_driver(struct spi_driver *);

extern int spi_attach_client(struct spi_client *);
extern int spi_detach_client(struct spi_client *);

/* New function: This is to get an spi_client-struct for controlling the
   client either by using spi_control-function or having the
   client-module export functions that can be used with the spi_client
   -struct. */
extern struct spi_client *spi_get_client(int driver_id, int adapter_id,
					struct spi_client *prev);

/* Should be used with new function
   extern struct spi_client *spi_get_client(int,int,struct spi_client *);
   to make sure that client-struct is valid and that it is okay to access
   the spi-client.
   returns -EACCES if client doesn't allow use (default)
   returns -EBUSY if client doesn't allow multiple use (default) and
   usage_count >0 */
extern int spi_use_client(struct spi_client *);
extern int spi_release_client(struct spi_client *);

/* call the spi_client->command() of all attached clients with
 * the given arguments
 */
extern void spi_clients_command(struct spi_adapter *adap,
				unsigned int cmd, void *arg);

/* returns -EBUSY if a client with the same name is attached on the adapter,
 * 0 if not. Note that the only other place at which this is called is within
 * spi_attach_client; so you can cheat by simply not registering.
 * Not recommended, of course!
 */
extern int spi_check_clientname (struct spi_adapter *adapter, char *clientname);

/* Detect function. It iterates over all possible addresses itself.
 * It will only call found_proc if some client is connected at the
 * specific address (unless a 'force' matched);
 */
extern int spi_probe(struct spi_adapter *adapter,
		struct spi_client_address_data *address_data,
		int (*found_proc) (struct spi_adapter *, int, int));

/* An ioctl like call to set div. parameters of the adapter.
 */
extern int spi_control(struct spi_client *,unsigned int, unsigned long);

/* This call returns a unique low identifier for each registered adapter,
 * or -1 if the adapter was not registered.
 */
extern int spi_adapter_id(struct spi_adapter *adap);
extern struct spi_adapter* spi_get_adapter(int id);
extern void spi_put_adapter(struct spi_adapter *adap);


/* Return the functionality mask */
extern u32 spi_get_functionality (struct spi_adapter *adap);

/* Return 1 if adapter supports everything we need, 0 if not. */
extern int spi_check_functionality (struct spi_adapter *adap, u32 func);

/*
 * SPI Message - used for pure spi transaction, also from /dev interface
 */
struct spi_msg {
 	__u16 flags;
#define SPI_M_CPOL_ACTHIGH	0x10
#define SPI_M_CPOL_ACTLOW	0x20
#define SPI_M_CPHA_FORMATA	0x40
#define SPI_M_CPHA_FORMATB	0x80
 	__u16 len;		/* msg length				*/
 	__u8 *wbuf;		/* pointer to msg data to write	*/
 	__u8 *rbuf;		/* pointer to msg data for read */
};

/* To determine what functionality is present */

#define SPI_FUNC_SPI			0x00000001
#define SPI_FUNC_CPOL			0x00000010
#define SPI_FUNC_CPHA			0x00000020

/* ----- commands for the ioctl like spi_command call:
 * note that additional calls are defined in the algorithm and hw
 *	dependent layers - these can be listed here, or see the
 *	corresponding header files.
 */
				/* -> bit-adapter specific ioctls	*/
#define SPI_RETRIES	0x0701	/* number of times a device address      */
				/* should be polled when not            */
                                /* acknowledging 			*/
#define SPI_TIMEOUT	0x0702	/* set timeout - call with int 		*/


/* this is for spi-dev.c	*/
#define SPI_FUNCS	0x0705	/* Get the adapter functionality */
#define SPI_XFER	0x0707	/* Combined transfer */

/* ----- SPI-DEV: char device interface stuff ------------------------- */

#define SPI_MAJOR	153		/* Device major number		*/
							/* minor 0-15 spi0 - spi15 */

/* These defines are used for probing spi client addresses */
/* Default fill of many variables */
#define SPI_CLIENT_DEFAULTS {SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END, \
                          SPI_CLIENT_END, SPI_CLIENT_END, SPI_CLIENT_END}

/* This is ugly. We need to evaluate SPI_CLIENT_MAX_OPTS before it is
   stringified */
#define SPI_CLIENT_MODPARM_AUX1(x) "1-" #x "h"
#define SPI_CLIENT_MODPARM_AUX(x) SPI_CLIENT_MODPARM_AUX1(x)
#define SPI_CLIENT_MODPARM SPI_CLIENT_MODPARM_AUX(SPI_CLIENT_MAX_OPTS)

/* SPI_CLIENT_MODULE_PARM creates a module parameter, and puts it in the
   module header */

#define SPI_CLIENT_MODULE_PARM(var,desc) \
  static unsigned short var[SPI_CLIENT_MAX_OPTS] = SPI_CLIENT_DEFAULTS; \
  static unsigned int var##_num; \
  /*MODULE_PARM(var,SPI_CLIENT_MODPARM);*/ \
  module_param_array(var, short, &var##_num, 0); \
  MODULE_PARM_DESC(var,desc)

/* This is the one you want to use in your own modules */
#define SPI_CLIENT_INSMOD \
  SPI_CLIENT_MODULE_PARM(probe, \
                      "List of adapter,address pairs to scan additionally"); \
  SPI_CLIENT_MODULE_PARM(probe_range, \
                      "List of adapter,start-addr,end-addr triples to scan " \
                      "additionally"); \
  SPI_CLIENT_MODULE_PARM(ignore, \
                      "List of adapter,address pairs not to scan"); \
  SPI_CLIENT_MODULE_PARM(ignore_range, \
                      "List of adapter,start-addr,end-addr triples not to " \
                      "scan"); \
  SPI_CLIENT_MODULE_PARM(force, \
                      "List of adapter,address pairs to boldly assume " \
                      "to be present"); \
	static struct spi_client_address_data addr_data = {		\
			.normal_spi = 		normal_spi,		\
			.normal_spi_range =	normal_spi_range,	\
			.probe =		probe,			\
			.probe_range =		probe_range,		\
			.ignore =		ignore,			\
			.ignore_range =		ignore_range,		\
			.force =		force,			\
		}

/* Detect whether we are on the isa bus. If this returns true, all spi
   access will fail! */
#define spi_is_isa_client(clientptr) \
        ((clientptr)->adapter->algo->id == SPI_ALGO_ISA)
#define spi_is_isa_adapter(adapptr) \
        ((adapptr)->algo->id == SPI_ALGO_ISA)

#endif /* _LINUX_SPI_H */
