/*
 * OHCI HCD (Host Controller Driver) for USB.
 *
 * (C) Copyright 1999 Roman Weissgaerber <weissg@vienna.at>
 * (C) Copyright 2000-2002 David Brownell <dbrownell@users.sourceforge.net>
 * (C) Copyright 2002 Hewlett-Packard Company
 *
 * Bus Glue for NS9360 and NS9750
 *
 * Written by Christopher Hoover <ch@hpl.hp.com>
 * Based on fragments of previous driver by Rusell King et al.
 *
 * Modified for NS9XXX from ohci-sa1111.c
 *  by  Jonas Dietsche <jdietsche@fsforth.de>
 *
 * This file is licenced under the GPL.
 */

#include <asm/hardware.h>
#include <asm/mach-types.h>
#include <asm/arch/regs-usb.h>
#include <asm/arch/regs-bbus.h>
#include <asm/arch/regs-bbus_bridge.h>
#include <asm/arch/irqs.h>
#include <asm/arch/processor.h>
#include <linux/platform_device.h>

// #define DEBUG_USB

#ifdef DEBUG_USB
# define DEBUG(output, ...)	do { printk( KERN_INFO "%s: " output "\n", __func__, ##__VA_ARGS__); } while( 0 )
# define DEBUG_FN()		do { printk( KERN_INFO "%s:line %d\n", __func__, __LINE__ ); } while( 0 )
#else
# define DEBUG(output, ...)
# define DEBUG_FN()
#endif

extern int usb_disabled(void);

/*-------------------------------------------------------------------------*/

static void ns9xxx_start_hc( struct platform_device *dev )
{
	DEBUG_FN();
	/* enable USB Host mode */
	/* XXX Use propper constants through header file */
	if (processor_is_ns9360()) {
		BBUS_MRST |= 0x800;
		BBUS_USB_CFG = NS_BBUS_USB_CFG_SPEED;
		BBUS_MRST &= ~0x800;
		UHFE_IEN = NS_USB_INT_OHCI_IRQ;
	} else if (processor_is_ns9750()) {
		printk( "processor is ns9750" );
		BBUS_MRST |= 0x002;
		BBUS_USB_CFG = NS_BBUS_USB_CFG_SPEED | NS_BBUS_USB_CFG_CFG_DEVICE;
		BBUS_MRST &= ~0x002;
		USB_GCTRL_STAT = 0x0;	/* Configure for host mode */
	}
}

static void ns9xxx_stop_hc(struct platform_device *dev)
{
	printk(KERN_DEBUG "%s: stopping NS9XXX OHCI USB Controller\n", dev->name);
}


/*-------------------------------------------------------------------------*/


static irqreturn_t usb_hcd_ns9xxx_hcim_irq (struct usb_hcd *hcd, struct pt_regs *ptregs)
{
	irqreturn_t ret = ohci_irq( hcd, ptregs );
	/* acknowledge interrupt */
#if defined(CONFIG_PROCESSOR_NS9360)
	UHFE_ISTAT |= NS_USB_INT_OHCI_IRQ;
#elif defined(CONFIG_PROCESSOR_NS9750)
	USB_ISTAT |= NS_USB_INT_OHCI_IRQ;
#endif
	return ret;
}

/*-------------------------------------------------------------------------*/

static int usb_hcd_ns9xxx_remove (struct usb_hcd *hcd, struct platform_device *pdev);


/* configure so an HC device and id are always provided */
/* always called with process context; sleeping is OK */



static int __devinit ohci_ns9xxx_start (struct usb_hcd *hcd)
{
	struct ohci_hcd	*ohci = hcd_to_ohci (hcd);
	int		ret;

	ohci_dbg (ohci, "ohci_nsxxx_start, ohci:%p", ohci);

	if ((ret = ohci_init(ohci)) < 0)
		return ret;

	if ((ret = ohci_run (ohci)) < 0) {
		err ("can't start %s", hcd->self.bus_name);
		ohci_stop (hcd);
		return ret;
	}

	return 0;
}


/**
 * usb_hcd_ns9xxx_probe - initialize NS9XXX-based HCDs
 * Context: !in_interrupt()
 *
 * Allocates basic resources for this USB host controller, and
 * then invokes the start() method for the HCD associated with it
 * through the hotplug entry's driver_data.
 *
 * Store this function in the HCD's struct pci_driver as probe().
 */
int usb_hcd_ns9xxx_probe (const struct hc_driver *driver, struct platform_device *pdev)
{
	int retval;
	struct usb_hcd *hcd = NULL;

	DEBUG_FN();

	if (pdev->num_resources != 2) {
		pr_debug("hcd probe: invalid num_resources");
		return -ENODEV;
	}

	if ((pdev->resource[0].flags != IORESOURCE_MEM) || (pdev->resource[1].flags != IORESOURCE_IRQ)) {
		pr_debug("hcd probe: invalid resource type\n");
		return -ENODEV;
	}

	hcd = usb_create_hcd(driver, &pdev->dev, "ns9xxx");
	if (!hcd)
		return -ENOMEM;

	hcd->rsrc_start = pdev->resource[0].start;
	hcd->rsrc_len = pdev->resource[0].end - pdev->resource[0].start + 1;

	if (!request_mem_region(hcd->rsrc_start, hcd->rsrc_len, hcd_name)) {
		printk( KERN_WARNING "Unable to request mem region, start = 0x%x len = 0x%x\n",
			(u32)hcd->rsrc_start, (u32)hcd->rsrc_len );
		retval = -EBUSY;
		goto err1;
	}

	hcd->regs = ioremap(hcd->rsrc_start + USB_REG_OFFSET, hcd->rsrc_len - USB_REG_OFFSET);
	if (!hcd->regs) {
		printk( KERN_WARNING "Unable to ioremap, start = 0x%x len = 0x%x\n",
			(u32)(hcd->rsrc_start + USB_REG_OFFSET), (u32)(hcd->rsrc_len - USB_REG_OFFSET) );
		retval = -EIO;
		goto err2;
	}

	ns9xxx_start_hc(pdev);
	ohci_hcd_init(hcd_to_ohci(hcd));

	retval = usb_add_hcd(hcd, pdev->resource[1].start, SA_INTERRUPT);
	if (retval != 0) {
		printk( KERN_WARNING "Unable to add hcd, ret = %d\n", retval );
		goto err3;
	}
	return 0;

err3:
	ns9xxx_stop_hc(pdev);
	iounmap(hcd->regs);
err2:
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
 err1:
	usb_put_hcd(hcd);
	return retval;
}


/* may be called without controller electrically present */
/* may be called with controller, bus, and devices active */

/**
 * usb_hcd_ns9xxx_remove - shutdown processing for NS9XXX-based HCDs
 * @dev: USB Host Controller being removed
 * Context: !in_interrupt()
 *
 * Reverses the effect of usb_hcd_ns9xxx_probe(), first invoking
 * the HCD's stop() method.  It is always called from a thread
 * context, normally "rmmod", "apmd", or something similar.
 *
 */
static int usb_hcd_ns9xxx_remove (struct usb_hcd *hcd, struct platform_device *pdev)
{
	usb_remove_hcd(hcd);
	ns9xxx_stop_hc(pdev);
	iounmap(hcd->regs);
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
	usb_put_hcd(hcd);

	return 0;
}


/*-------------------------------------------------------------------------*/

static const struct hc_driver ohci_ns9xxx_hc_driver = {
	.description =		hcd_name,
	.product_desc =		"NS9XXX OHCI",
	.hcd_priv_size =	sizeof(struct ohci_hcd),

	/*
	 * generic hardware linkage
	 */
	.irq =			usb_hcd_ns9xxx_hcim_irq,
	.flags =		HCD_USB11 | HCD_MEMORY,

	/*
	 * basic lifecycle operations
	 */
	.start =		ohci_ns9xxx_start,
	.stop =			ohci_stop,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue =		ohci_urb_enqueue,
	.urb_dequeue =		ohci_urb_dequeue,
	.endpoint_disable =	ohci_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number =	ohci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data =	ohci_hub_status_data,
	.hub_control =		ohci_hub_control,
#ifdef	CONFIG_PM
	.bus_suspend =		ohci_bus_suspend,
	.bus_resume =		ohci_bus_resume,
#endif
	.start_port_reset =	ohci_start_port_reset,
};


static int ohci_hcd_ns9xxx_drv_probe(struct platform_device *dev)
{
	return usb_hcd_ns9xxx_probe(&ohci_ns9xxx_hc_driver, dev);
}

static int ohci_hcd_ns9xxx_drv_remove(struct platform_device *dev)
{
	return usb_hcd_ns9xxx_remove(platform_get_drvdata(dev), dev);
}


/* TODO XXX implement if PM required */
#define ohci_hcd_ns9xxx_drv_suspend NULL
#define ohci_hcd_ns9xxx_drv_resume  NULL


static struct platform_driver ohci_hcd_ns9xxx_driver = {
	.probe		= ohci_hcd_ns9xxx_drv_probe,
	.remove		= ohci_hcd_ns9xxx_drv_remove,
	.suspend	= ohci_hcd_ns9xxx_drv_suspend,
	.resume		= ohci_hcd_ns9xxx_drv_resume,
	.driver		= {
		.name	= "ns9xxx-ohci",
		.owner	= THIS_MODULE,
	},
};


static int __init ohci_hcd_ns9xxx_init (void)
{
	if (usb_disabled())
		return -ENODEV;

	return platform_driver_register(&ohci_hcd_ns9xxx_driver);
}

static void __exit ohci_hcd_ns9xxx_cleanup (void)
{
	platform_driver_unregister(&ohci_hcd_ns9xxx_driver);
}

module_init (ohci_hcd_ns9xxx_init);
module_exit (ohci_hcd_ns9xxx_cleanup);
