/*
 *  cmd_nvram/nvram_priv_uboot.c
 *
 *  Copyright (C) 2006 by Digi International Inc.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License version2  as published by
 *  the Free Software Foundation.
 */
/*
 *  !Revision:   $Revision: 1.13 $
 *  !Author:     Markus Pietrek
 *  !Descr:      Defines the private functions needed by the nvram core to
 *               access I2C, Flash and a console for U-Boot.
 */

#include <common.h>

#if (CONFIG_COMMANDS & CFG_CMD_NAND)
# include <nand.h>
#endif  /* CFG_CMD_NAND */

#include "nvram_priv.h"
#include "env.h"
#include "partition.h"

#define OFFS( x ) ( NAND_UBOOT_SIZE + ( x ) )

#define CE( sCmd ) \
        do { \
                if( !(sCmd) )       \
                        goto error;    \
        } while( 0 )

#if !defined( NAND_FPGA_SIZE )
# define NAND_FPGA_SIZE 0
#endif

DECLARE_GLOBAL_DATA_PTR;

/* ********** local functions ********** */

static int NvPrivOSIntInRange( loff_t iOffs, size_t iLength );

/* ********** local variables ********** */

static nand_info_t* pChip = &nand_info[ 0 ];

/* ********** global functions ********** */

int NvPrivOSInit( void )
{
        return 1;
}

int NvPrivOSPostInit( void )
{
        return 1;
}

int NvPrivOSFinish( void )
{
        return 1;
}

int NvPrivOSCriticalPostReset(
        struct nv_critical* pCrit )
{
        CE( NvOSCfgAdd( NVOS_UBOOT, CFG_ENV_SIZE ) );

        /* nothing to do */
        return 1;

error:
        return 0;
}

int NvPrivOSCriticalPartReset(
        struct nv_critical* pCrit,
        nv_os_type_e eForOS )
{
        static const nv_param_part_t axPartitions[] = {
                { .szName = "U-Boot", .eType = NVPT_UBOOT, .flags.bFixed = 1,
                  .ullStart = 0, .ullSize = NAND_UBOOT_SIZE },
                { .szName = "NVRAM", .eType = NVPT_NVRAM, .flags.bFixed = 1,
                  .ullStart  = NAND_UBOOT_SIZE, .ullSize = NAND_NVRAM_SIZE },
#if NAND_FPGA_SIZE > 0
                { .szName = "FPGA", .eType = NVPT_FPGA, .flags.bFixed = 1,
                  .ullStart  = NAND_UBOOT_SIZE + NAND_NVRAM_SIZE,
                  .ullSize = NAND_FPGA_SIZE },
#endif
        };
        static const nv_param_part_t axPartitionsLinux[] = {
                { .szName = "Kernel", .eType = NVPT_LINUX, .flags.bFixed = 0,
                  .ullStart = NAND_UBOOT_SIZE + NAND_NVRAM_SIZE + NAND_FPGA_SIZE,
                  .ullSize  = NAND_KERNEL_SIZE },
                { .szName = "RootFS-JFFS2", .eType = NVPT_FILESYSTEM,
                  .ullStart = NAND_UBOOT_SIZE + NAND_NVRAM_SIZE + NAND_FPGA_SIZE + NAND_KERNEL_SIZE,
                  .ullSize  = NAND_ROOTFS_SIZE,
                  .flags.bFixed = 0, .flags.fs.eType = NVFS_JFFS2,
                  .flags.fs.bRoot = 1 },
                { .szName = "User-JFFS2", .eType = NVPT_FILESYSTEM,
                  .ullStart = NAND_UBOOT_SIZE + NAND_NVRAM_SIZE + NAND_FPGA_SIZE + NAND_KERNEL_SIZE + NAND_ROOTFS_SIZE,
                  .ullSize  = 0,
                  .flags.bFixed = 0, .flags.fs.eType = NVFS_JFFS2 },
        };
        static const nv_param_part_t axPartitionsWinCE[] = {
                { .szName = "EBoot", .eType = NVPT_EBOOT, .flags.bFixed = 0,
                  .ullStart = NAND_UBOOT_SIZE + NAND_NVRAM_SIZE + NAND_FPGA_SIZE,
                  .ullSize  = NAND_EBOOT_SIZE },
                { .szName = "Registry", .eType = NVPT_WINCE_REGISTRY,
                  .ullStart = NAND_UBOOT_SIZE + NAND_NVRAM_SIZE + NAND_FPGA_SIZE + NAND_EBOOT_SIZE,
                  .ullSize  = NAND_WINCE_REG_SIZE,
                  .flags.bFixed = 0, .flags.fs.eType = NVFS_NONE,
                  .flags.fs.bRoot = 0 },
                { .szName = "Kernel", .eType = NVPT_WINCE,
                  .ullStart = NAND_UBOOT_SIZE + NAND_NVRAM_SIZE + NAND_FPGA_SIZE + NAND_EBOOT_SIZE + NAND_WINCE_REG_SIZE,
                  .ullSize  = NAND_WINCE_SIZE,
                  .flags.bFixed = 0, .flags.fs.eType = NVFS_NONE },
        };
        static const nv_param_part_t axPartitionsNetOS[] = {
                { .szName = "NET+OS-Kernel", .eType = NVPT_NETOS, .flags.bFixed = 0,
                  .ullStart = NAND_UBOOT_SIZE + NAND_NVRAM_SIZE + NAND_FPGA_SIZE,
                  .ullSize  = NAND_NETOS_KERNEL_SIZE },
                { .szName = "NET+OS-FS", .eType = NVPT_FILESYSTEM,
                  .ullStart = NAND_UBOOT_SIZE + NAND_NVRAM_SIZE + NAND_FPGA_SIZE + NAND_NETOS_KERNEL_SIZE,
                  .ullSize  = 0,
                  .flags.bFixed   = 0, .flags.fs.eType = NVFS_YAFFS,
                  .flags.fs.bRoot = 0 },
        };
        const nv_param_part_t* pOS = NULL;
        int iOSSize = 0;
        int i;
        
        for( i = 0; i < ARRAY_SIZE( axPartitions ); i++ ) {
                pCrit->s.p.xPartTable.uiEntries++;
                ASSERT( i < ARRAY_SIZE( pCrit->s.p.xPartTable.axEntries ) );
                pCrit->s.p.xPartTable.axEntries[ i ] = axPartitions[ i ];
        }

        switch( eForOS ) {
            case NVOS_NONE:
            case NVOS_UBOOT:
                /* U-Boot is always added, we need it for Update and NVRAM */
                /* we already added it */
                break;
            case NVOS_LINUX:
                pOS     = axPartitionsLinux;
                iOSSize = ARRAY_SIZE( axPartitionsLinux );
                break;
            case NVOS_WINCE:
                pOS     = axPartitionsWinCE;
                iOSSize = ARRAY_SIZE( axPartitionsWinCE );
                break;
            case NVOS_NETOS:
                pOS     = axPartitionsNetOS;
                iOSSize = ARRAY_SIZE( axPartitionsNetOS );
                break;
            default:
                NV_SET_ERROR( NVE_NOT_IMPLEMENTED, NvToStringOS( eForOS ) );
                break;
        }

        /* append OS stuff */
        for( i = 0; i < iOSSize; i++ ) {
                nv_param_part_t* pPart = &pCrit->s.p.xPartTable.axEntries[ pCrit->s.p.xPartTable.uiEntries ];
                ASSERT( i < ARRAY_SIZE( pCrit->s.p.xPartTable.axEntries ) - 1 );
                *pPart = pOS[ i ];
                if( !pPart->ullSize )
                        /* convert 0 to to-end-of-flash */
                        pPart->ullSize = PartSize( pPart );

                pCrit->s.p.xPartTable.uiEntries++;
        }

        return 1;
}

int NvPrivOSI2CPresent( void )
{
        /* !TODO */
        return 0;
}

/* !TODO deprecate to warn not implemented */
int NvPrivOSI2COpen( char bForWrite )
{
        RETURN_NOT_IMPLEMENTED();
}

int NvPrivOSI2CClose( void )
{
        RETURN_NOT_IMPLEMENTED();
}

int NvPrivOSI2CRead( void* pvbuf, size_t iLength )
{
        RETURN_NOT_IMPLEMENTED();
}

int NvPrivOSI2CdWrite( const void* pvbuf, size_t iLength )
{
        RETURN_NOT_IMPLEMENTED();
}

int NvPrivOSFlashOpen( char bForWrite )
{
        /* not necessary */
        return 1;
}

int NvPrivOSFlashClose( void )
{
        /* not necessary */
        return 1;
}

int NvPrivOSFlashRead( void* pvBuf, loff_t iOffs, size_t iLength )
{
        size_t iRetBlock;
        
        CE( NvPrivOSIntInRange( iOffs, iLength ) );
        if( pChip->read( pChip, OFFS( iOffs ), iLength, &iRetBlock, pvBuf ) ) {
                char szErr[ 64 ];
                sprintf( szErr, "Flash Read @ 0x%08qx, length = %i",
                         OFFS( iOffs ), iLength );
                
                NV_SET_ERROR( NVE_IO, szErr );
                goto error;
        }
        
        return 1;

error:
        return 0;
}

int NvPrivOSFlashErase( loff_t iOffs )
{
        struct erase_info xErase;

        CLEAR( xErase );

	xErase.mtd  = pChip;
	xErase.addr = OFFS( iOffs );
	xErase.len  = pChip->erasesize;
	xErase.callback = 0;
        
        CE( NvPrivOSIntInRange( iOffs, 0 ) );
        if( pChip->erase( pChip, &xErase ) ) {
                char szErr[ 64 ];
                sprintf( szErr, "Flash Erase @ 0x%08qx, length = %i",
                         OFFS( iOffs ), xErase.len );
                
                goto error;
        }

        return 1;

error:
        return 0;
}

int NvPrivOSFlashWrite( /*@in@*/ const void* pvBuf, loff_t iOffs, size_t iLength )
{
        size_t iRetBlock;
        
        CE( NvPrivOSIntInRange( iOffs, iLength ) );
        if( pChip->write( pChip, OFFS( iOffs ), iLength, &iRetBlock,
                          pvBuf ) ) {
                char szErr[ 64 ];
                sprintf( szErr, "Flash Write @ 0x%08qx, length = %i",
                         OFFS( iOffs ), iLength );
                
                NV_SET_ERROR( NVE_IO, szErr );
                goto error;
        }
        

        return 1;

error:
        return 0;
}

int NvPrivOSFlashInfo(
        loff_t iOffs,
        /*@out@*/ struct nv_priv_flash_status* pStatus )
{
        if( !NvPrivOSIntInRange( iOffs, 0 ) )
                return 0;
        
        CLEAR( *pStatus );
        pStatus->bBad       = pChip->block_isbad( pChip, OFFS( iOffs ) );
        pStatus->iEraseSize = pChip->erasesize;

        return 1;
}

void NvPrivOSPrintf( const char* szFormat, ...)
{
        va_list ap;

        if( !NvPrivIsOutputEnabled() )
                return;
        
        va_start( ap, szFormat );
        vprintf( szFormat, ap );
        va_end( ap );
}

void NvPrivOSPrintfError( const char* szFormat, ...)
{
        va_list ap;

        if( !NvPrivIsOutputEnabled() )
                return;
        
        /* it's not stderr, put there is no vfprintf( stderr )*/
        va_start( ap, szFormat );
        vprintf( szFormat, ap );
        va_end( ap );
}

static int NvPrivOSIntInRange( loff_t iOffs, size_t iLength )
{
        int iRes = 1;
        
        if( ( iOffs + iLength ) > NAND_NVRAM_SIZE )
                iRes = NV_SET_ERROR( NVE_NO_SPACE, "" );

        return iRes;
}
